from collections.abc import Mapping as C_Mapping, Sequence as C_Sequence, Callable as C_Callable
from datetime import datetime, date  # noqa: I251
import inspect
import os
import sys
from re import Pattern as _REPattern
from types import FunctionType, ModuleType
from typing import (
    Callable,
    ClassVar,
    Dict,
    Any,
    Final,
    Literal,
    List,
    Mapping,
    NewType,
    Optional,
    Tuple,
    Type,
    Protocol,
    TYPE_CHECKING,
    Union,
    runtime_checkable,
    IO,
    Iterator,
    Generator,
    NamedTuple,
    Sequence,
)

from typing_extensions import (
    ForwardRef,
    Annotated,
    Never,
    ParamSpec,
    TypeAlias,
    Concatenate,
    Unpack,
    Self,
    Generic,
    get_args,
    TypeVar,
    get_origin,
    get_type_hints,
    get_origin,
    get_original_bases,
)

from typing_extensions import is_typeddict as _is_typeddict

from typing_extensions import TypedDict  # noqa: I251

try:
    from types import UnionType  # type: ignore[attr-defined]
except ImportError:
    # Since new Union syntax was introduced in Python 3.10
    # we need to substitute it here for older versions.
    # it is defined as type(int | str) but for us having it
    # as shown here should suffice because it is valid only
    # in versions of Python>=3.10.
    UnionType = Never

from typing import _SpecialGenericAlias, _GenericAlias  # type: ignore[attr-defined]
from types import GenericAlias

typingGenericAlias: Tuple[Any, ...] = (_GenericAlias, _SpecialGenericAlias, GenericAlias)


from dlt.common.pendulum import timedelta, pendulum

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath
    from typing import _TypedDict

    REPattern = _REPattern[str]
    PathLike = os.PathLike[str]
else:
    StrOrBytesPath = Any
    from typing import _TypedDictMeta as _TypedDict

    REPattern = _REPattern
    PathLike = os.PathLike


AnyType: TypeAlias = Any
CallableAny = NewType("CallableAny", Any)  # type: ignore[valid-newtype]
"""A special callable Any that returns argument but is recognized as Any type by dlt hint checkers"""
NoneType = type(None)
DictStrAny: TypeAlias = Dict[str, Any]
DictStrStr: TypeAlias = Dict[str, str]
DictStrOptionalStr: TypeAlias = Dict[str, Optional[str]]
StrAny: TypeAlias = Mapping[str, Any]  # immutable, covariant entity
StrStr: TypeAlias = Mapping[str, str]  # immutable, covariant entity
StrStrStr: TypeAlias = Mapping[str, Mapping[str, str]]  # immutable, covariant entity
AnyFun: TypeAlias = Callable[..., Any]
TFun = TypeVar("TFun", bound=AnyFun)  # any function
TAny = TypeVar("TAny", bound=Any)
TAnyFunOrGenerator = TypeVar(
    "TAnyFunOrGenerator", AnyFun, Generator[Any, Optional[Any], Optional[Any]]
)
TAnyClass = TypeVar("TAnyClass", bound=object)
TimedeltaSeconds = Union[int, float, timedelta]
# represent secret value ie. coming from Kubernetes/Docker secrets or other providers


class SecretSentinel:
    """Marks a secret type when part of type annotations"""


if TYPE_CHECKING:
    TSecretValue = Annotated[Any, SecretSentinel]
else:
    # use callable Any type for backward compatibility at runtime
    TSecretValue = Annotated[CallableAny, SecretSentinel]

TSecretStrValue = Annotated[str, SecretSentinel]

TColumnNames = Union[str, Sequence[str]]
"""A string representing a column name or a sequence of"""
TTableNames = Union[str, Sequence[str]]
"""A string representing a table name or a sequence of"""
TDataItem: TypeAlias = Any
"""A single data item as extracted from data source"""
TDataItems: TypeAlias = Union[TDataItem, List[TDataItem]]
"A single data item or a list as extracted from the data source"
TAnyDateTime = Union[pendulum.DateTime, pendulum.Date, datetime, date, str, float, int]
"""DateTime represented as pendulum/python object, ISO string or unix timestamp"""
TVariantBase = TypeVar("TVariantBase", covariant=True)
TVariantRV = Tuple[str, Any]
VARIANT_FIELD_FORMAT = "v_%s"
TFileOrPath = Union[str, PathLike, IO[Any]]
TSortOrder = Literal["asc", "desc"]
TLoaderFileFormat = Literal[
    "jsonl", "typed-jsonl", "insert_values", "parquet", "csv", "reference", "model"
]
"""known loader file formats"""

TDynHintType = TypeVar("TDynHintType")
TFunHintTemplate = Callable[[TDataItem], TDynHintType]
TTableHintTemplate = Union[TDynHintType, TFunHintTemplate[TDynHintType]]


class ConfigValueSentinel(NamedTuple):
    """Class to create singleton sentinel for config and secret injected value"""

    default_literal: str
    default_type: AnyType

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        if self.default_literal == "dlt.config.value":
            inst_ = "ConfigValue"
        else:
            inst_ = "SecretValue"
        return f"{inst_}({self.default_literal}) awaiting injection"


ConfigValue: None = ConfigValueSentinel("dlt.config.value", AnyType)  # type: ignore[assignment]
"""Config value indicating argument that may be injected by config provider. Evaluates to None when type checking"""

SecretValue: None = ConfigValueSentinel("dlt.secrets.value", TSecretValue)  # type: ignore[assignment]
"""Secret value indicating argument that may be injected by config provider. Evaluates to None when type checking"""


@runtime_checkable
class SupportsVariant(Protocol, Generic[TVariantBase]):
    """Defines variant type protocol that should be recognized by normalizers

    Variant types behave like TVariantBase type (ie. Decimal) but also implement the protocol below that is used to extract the variant value from it.
    See `Wei` type declaration which returns Decimal or str for values greater than supported by destination warehouse.
    """

    def __call__(self) -> Union[TVariantBase, TVariantRV]: ...


class SupportsHumanize(Protocol):
    def asdict(self) -> DictStrAny:
        """Represents object as dict with a schema loadable by dlt"""
        ...

    def asstr(self, verbosity: int = 0) -> str:
        """Represents object as human readable string"""
        ...


class SocketLike(Protocol):
    def send(self, data: bytes) -> int:
        """Send bytes; return number of bytes sent."""
        ...

    def recv(self, bufsize: int) -> bytes:
        """Receive up to `bufsize` bytes and return them."""
        ...

    def close(self) -> None:
        """Close the socket-like connection."""
        ...

    def settimeout(self, timeout: float) -> None:
        """Set the timeout (in seconds) for send/recv operations."""
        ...


def get_type_name(t: Type[Any]) -> str:
    """Returns a human-friendly name of type `t`"""
    if name := getattr(t, "__name__", None):
        return name  # type: ignore[no-any-return]
    t = get_origin(t) or t
    if name := getattr(t, "__name__", None):
        return name  # type: ignore[no-any-return]
    return str(t)


def is_callable_type(hint: Type[Any]) -> bool:
    """Checks if `hint` is callable: a function or callable class. This function does not descend
    into type arguments ie. if Union, Literal or NewType contain callables, those are ignored"""
    if get_origin(hint) is get_origin(Callable):
        return True
    # this skips NewType etc.
    if getattr(hint, "__module__", None) == "typing":
        return False
    if isinstance(hint, FunctionType):
        return True
    # this is how we check if __call__ is implemented in the mro
    if inspect.isclass(hint) and any("__call__" in t_.__dict__ for t_ in inspect.getmro(hint)):
        return True

    return False


def extract_type_if_modifier(t: Type[Any], preserve_annotated: bool = False) -> Optional[Type[Any]]:
    modifiers = (Final, ClassVar) if preserve_annotated else (Final, ClassVar, Annotated)
    if get_origin(t) in modifiers:
        t = get_args(t)[0]
        if m_t := extract_type_if_modifier(t):
            return m_t
        else:
            return t
    return None


def extract_supertype(t: Type[Any]) -> Optional[Type[Any]]:
    return getattr(t, "__supertype__", None)  # type: ignore[no-any-return]


def is_union_type(hint: Type[Any]) -> bool:
    # We need to handle UnionType because with Python>=3.10
    # new Optional syntax was introduced which treats Optionals
    # as unions and probably internally there is no additional
    # type hints to handle this edge case, see the examples below
    # >>> type(str | int)
    # <class 'types.UnionType'>
    # >>> type(str | None)
    # <class 'types.UnionType'>
    # type(Union[int, str])
    # <class 'typing._GenericAlias'>
    origin = get_origin(hint)
    if origin is Union or origin is UnionType:
        return True

    if inner_t := extract_type_if_modifier(hint):
        return is_union_type(inner_t)

    return False


def is_any_type(t: Type[Any]) -> bool:
    """Checks if `t` is one of recognized Any types"""
    return t in (Any, CallableAny)


def is_optional_type(t: Type[Any]) -> bool:
    origin = get_origin(t)
    is_union = origin is Union or origin is UnionType
    if is_union and type(None) in get_args(t):
        return True

    if inner_t := extract_type_if_modifier(t):
        if is_optional_type(inner_t):
            return True
        else:
            t = inner_t
    if super_t := extract_supertype(t):
        return is_optional_type(super_t)

    return False


def is_final_type(t: Type[Any]) -> bool:
    return get_origin(t) is Final


def extract_union_types(t: Type[Any], no_none: bool = False) -> List[Any]:
    if no_none:
        return [arg for arg in get_args(t) if arg is not type(None)]  # noqa: E721
    return list(get_args(t))


def is_literal_type(hint: Type[Any]) -> bool:
    if get_origin(hint) is Literal:
        return True
    if inner_t := extract_type_if_modifier(hint):
        if is_literal_type(inner_t):
            return True
        else:
            hint = inner_t
    if super_t := extract_supertype(hint):
        return is_literal_type(super_t)
    if is_union_type(hint) and is_optional_type(hint):
        return is_literal_type(get_args(hint)[0])

    return False


def get_literal_args(literal: Type[Any]) -> List[Any]:
    """Recursively get arguments from nested Literal types and return an unified list."""
    if not hasattr(literal, "__origin__") or literal.__origin__ is not Literal:
        raise ValueError(f"Provided object is not a `Literal`. Object: {literal}")

    unified_args = []

    def _get_args(literal: Type[Any]) -> None:
        for arg in get_args(literal):
            if hasattr(arg, "__origin__") and arg.__origin__ is Literal:
                _get_args(arg)
            else:
                unified_args.append(arg)

    _get_args(literal)

    return unified_args


def is_newtype_type(t: Type[Any]) -> bool:
    if hasattr(t, "__supertype__"):
        return True
    if inner_t := extract_type_if_modifier(t):
        if is_newtype_type(inner_t):
            return True
        else:
            t = inner_t
    if is_union_type(t) and is_optional_type(t):
        return is_newtype_type(get_args(t)[0])
    return False


def is_typeddict(t: Type[Any]) -> bool:
    if _is_typeddict(t):
        return True
    if inner_t := extract_type_if_modifier(t):
        return is_typeddict(inner_t)
    return False


def is_annotated(ann_type: Any) -> bool:
    try:
        return get_origin(ann_type) is Annotated
    except TypeError:
        return False


def is_list_generic_type(t: Type[Any]) -> bool:
    try:
        return issubclass(get_origin(t), C_Sequence)
    except TypeError:
        return False


def is_dict_generic_type(t: Type[Any]) -> bool:
    try:
        return issubclass(get_origin(t), C_Mapping)
    except TypeError:
        return False


def extract_inner_type(
    hint: Type[Any],
    preserve_new_types: bool = False,
    preserve_literal: bool = False,
    preserve_annotated: bool = False,
) -> Type[Any]:
    """Gets the inner type from Literal, Optional, Final and NewType

    Args:
        hint (Type[Any]): Type to extract
        preserve_new_types (bool): Do not extract supertype of a NewType

    Returns:
        Type[Any]: Inner type if hint was Literal, Optional or NewType, otherwise hint
    """
    if maybe_modified := extract_type_if_modifier(hint, preserve_annotated):
        return extract_inner_type(
            maybe_modified, preserve_new_types, preserve_literal, preserve_annotated
        )
    # make sure we deal with optional directly
    if is_union_type(hint) and is_optional_type(hint):
        return extract_inner_type(
            get_args(hint)[0], preserve_new_types, preserve_literal, preserve_annotated
        )
    if is_literal_type(hint) and not preserve_literal:
        # assume that all literals are of the same type
        return type(get_args(hint)[0])
    if hasattr(hint, "__supertype__") and not preserve_new_types:
        # descend into supertypes of NewType
        return extract_inner_type(
            hint.__supertype__, preserve_new_types, preserve_literal, preserve_annotated
        )
    return hint


def get_all_types_of_class_in_union(
    hint: Any, cls: TAny, with_superclass: bool = False
) -> List[TAny]:
    """if `hint` is an Union that contains classes, return all classes that are a subclass or (optionally) superclass of cls"""
    return [
        t
        for t in get_args(hint)
        if not is_typeddict(t) and (is_subclass(t, cls) or is_subclass(cls, t) and with_superclass)
    ]


def is_generic_alias(tp: Any) -> bool:
    """Tests if type is a generic alias ie. List[str]"""
    return isinstance(tp, typingGenericAlias) and tp.__origin__ not in (
        Union,
        tuple,
        ClassVar,
        C_Callable,
    )


def is_subclass(subclass: Any, cls: Any) -> bool:
    """Return whether 'cls' is a derived from another class or is the same class.

    Will handle generic types by comparing their origins.
    """
    if is_generic_alias(subclass):
        subclass = get_origin(subclass)
    if is_generic_alias(cls):
        cls = get_origin(cls)

    if inspect.isclass(subclass) and inspect.isclass(cls):
        return issubclass(subclass, cls)
    return False


def get_generic_type_argument_from_instance(
    instance: Any, sample_value: Optional[Any] = None
) -> Type[Any]:
    """Infers type argument of a Generic class from an `instance` of that class using optional `sample_value` of the argument type

    Inference depends on the presence of __orig_class__ attribute in instance, if not present - sample_Value will be used

    Args:
        instance (Any): instance of Generic class
        sample_value (Optional[Any]): instance of type of generic class, optional

    Returns:
        Type[Any]: type argument or Any if not known
    """
    orig_param_type = Any
    if cls_ := getattr(instance, "__orig_class__", None):
        cls_ = extract_inner_type(cls_)
    elif bases_ := get_original_bases(instance.__class__):
        # instance of class deriving from generic
        cls_ = bases_[0]
    if cls_:
        orig_param_type = get_args(cls_)[0]
    if orig_param_type in (Any, CallableAny) and sample_value is not None:
        orig_param_type = type(sample_value)  # type: ignore[assignment]
    return orig_param_type  # type: ignore


TInputArgs = ParamSpec("TInputArgs")
TReturnVal = TypeVar("TReturnVal")


def copy_sig(
    wrapper: Callable[TInputArgs, Any],
) -> Callable[[Callable[..., TReturnVal]], Callable[TInputArgs, TReturnVal]]:
    """Copies docstring and signature from wrapper to func but keeps the func return value type"""

    def decorator(func: Callable[..., TReturnVal]) -> Callable[TInputArgs, TReturnVal]:
        func.__doc__ = wrapper.__doc__
        return func

    return decorator


def copy_sig_ret(
    wrapper: Callable[TInputArgs, Any],
    ret: Type[TReturnVal],
) -> Callable[[Callable[TInputArgs, Any]], Callable[TInputArgs, TReturnVal]]:
    """Copies docstring and signature from wrapper to func and replaces return type with `ret`"""

    def decorator(func: Callable[TInputArgs, Any]) -> Callable[TInputArgs, TReturnVal]:
        func.__doc__ = wrapper.__doc__
        return func

    return decorator


def copy_sig_any(
    wrapper: Callable[Concatenate[TDataItem, TInputArgs], Any],
) -> Callable[
    [Callable[..., TReturnVal]], Callable[Concatenate[TDataItem, TInputArgs], TReturnVal]
]:
    """Copies docstring and signature from wrapper to func but keeps the func return value type

    It converts the type of first argument of the wrapper to Any which allows to type transformers in DltSources.
    See filesystem source readers as example
    """

    def decorator(
        func: Callable[..., TReturnVal]
    ) -> Callable[Concatenate[Any, TInputArgs], TReturnVal]:
        func.__doc__ = wrapper.__doc__
        return func

    return decorator


def add_value_to_literal(literal: Any, value: Any) -> None:
    """Extends a Literal at runtime with a new value.

    Args:
        literal (Type[Any]): Literal to extend
        value (Any): Value to add

    """
    type_args = get_args(literal)

    if value not in type_args:
        type_args += (value,)
        literal.__args__ = type_args


def get_type_globals(obj: Any) -> Dict[str, Any]:
    """
    Best-effort extraction of globals() associated with a type. If object is passed,
    we get its __class__

    Handles:
      - functions
      - classes (including TypedDict, dataclasses, Pydantic models, etc.)
      - modules (returns their __dict__)
    """

    # 1. Module: just return its dict
    if isinstance(obj, ModuleType):
        return obj.__dict__

    # 2. Function or bound/unbound method
    if inspect.isfunction(obj):
        return obj.__globals__

    # 3. Class (includes TypedDict, dataclasses, normal classes, etc.)
    if not inspect.isclass(obj):
        obj = obj.__class__

    if mod := sys.modules.get(obj.__module__):
        return mod.__dict__
    return {}


def resolve_single_annotation(
    ann: Any,
    *,
    globalns: Optional[Dict[str, Any]] = None,
    localns: Optional[Dict[str, Any]] = None,
    raise_on_error: bool = False,
) -> Any:
    """
    Resolves annotation `ann` if it is a str and/or ForwardRef.
    - If `ann` is not a str or ForwardRef, it's returned unchanged.
    - If it *is* a str/ForwardRef, we eval it in an appropriate namespace.
    """

    # fast path: already a real type
    if not isinstance(ann, (str, ForwardRef)):
        return ann

    # extract the expression and module from ForwardRef if needed
    expr: str
    if isinstance(ann, ForwardRef):
        expr = ann.__forward_arg__
        if (
            module := sys.modules.get(getattr(ann, "__forward_module__", None))
        ) and globalns is None:
            globalns = module.__dict__
    else:
        expr = ann

    try:
        ann = eval(expr, globalns, localns)
        if isinstance(ann, ForwardRef):
            ann = resolve_single_annotation(
                ann, globalns=globalns, localns=localns, raise_on_error=raise_on_error
            )
    except Exception:
        if raise_on_error:
            raise

    return ann



from typing import (
    List,
    Literal,
)
from dlt.common.typing import TypedDict

TExecInfoNames = Literal[
    "kubernetes",
    "docker",
    "codespaces",
    "github_actions",
    "airflow",
    "notebook",
    "colab",
    "aws_lambda",
    "gcp_cloud_function",
    "streamlit",
    "dagster",
    "prefect",
    "marimo",
    "claude_code",
    "claude_code_cli",
    "cursor",
    "codex",
]


class TVersion(TypedDict):
    """TypeDict representing a library version"""

    name: str
    version: str


class TExecutionContext(TypedDict, total=False):
    """TypeDict representing the runtime context info"""

    ci_run: bool
    python: str
    cpu: int
    exec_info: List[TExecInfoNames]
    library: TVersion
    os: TVersion
    run_context: str
    dlthub: TVersion



import hashlib
import os
import threading
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import tomlkit
import yaml

from dlt.common.json import json
from dlt.common.pendulum import pendulum
from dlt.common.utils import uniq_id, update_dict_nested
from dlt.common.configuration.const import TYPE_EXAMPLES
from dlt.common.configuration.providers.toml import SecretsTomlProvider, SettingsTomlProvider
from dlt.common.pipeline import get_dlt_repos_dir
from dlt.version import __version__ as dlt_ver

from dlt._workspace.typing import TWorkbenchComponentInfo
from dlt._workspace.cli.formatters import (
    extract_first_heading,
    parse_frontmatter,
    render_frontmatter,
    read_md_name_desc,
)
from dlt._workspace.cli.utils import get_provider_locations, make_dlt_settings_path
from dlt._workspace.typing import (
    TAiStatusInfo,
    TAiStatusWarning,
    TLocationInfo,
    TToolkitIndexEntry,
    TToolkitInfo,
    TWorkbenchToolkitInfo,
)
from dlt._workspace.cli._urls import DEFAULT_AI_WORKBENCH_BRANCH  # noqa: F401
from dlt._workspace.cli._urls import DEFAULT_AI_WORKBENCH_REPO  # noqa: F401

AI_WORKBENCH_BASE_DIR = "workbench"
TOOLKITS_INDEX_FILE = ".toolkits"

_workbench_lock = threading.Lock()  # lock git clone operation


def compute_file_hash(file_path: Path) -> str:
    """Return the SHA3-256 hex digest of a file's raw bytes."""
    return hashlib.sha3_256(file_path.read_bytes()).hexdigest()


def compute_content_hash(content: str) -> str:
    """Return the SHA3-256 hex digest of a string encoded as UTF-8."""
    return hashlib.sha3_256(content.encode("utf-8")).hexdigest()


def home_dir() -> Optional[Path]:
    """Return the user home directory or None when unavailable."""
    home = os.path.expanduser("~")
    if home and home != "~" and os.path.isdir(home):
        return Path(home)
    return None


def _strip_mcp_type(servers: Dict[str, Any]) -> Dict[str, Any]:
    """Strip the `type` field from each server config (Claude -> other platform)."""
    return {name: {k: v for k, v in cfg.items() if k != "type"} for name, cfg in servers.items()}


def parse_json_mcp(content: str, top_key: str) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Parse a JSON MCP config file into (full_data, servers_dict)."""
    if not content.strip():
        return {}, {}
    data: Dict[str, Any] = json.loads(content)
    servers: Dict[str, Any] = data.get(top_key, {})
    return data, servers


def merge_json_mcp_servers(
    existing_content: str,
    new_servers: Dict[str, Any],
    top_key: str,
    strip_type: bool,
) -> str:
    """Merge new MCP server entries into a JSON config file."""
    data, servers = parse_json_mcp(existing_content, top_key)
    if not servers:
        servers = data.setdefault(top_key, {})
    servers.update(_strip_mcp_type(new_servers) if strip_type else new_servers)
    return json.dumps(data, pretty=True) + "\n"


def parse_toml_mcp(content: str, top_key: str) -> Tuple[tomlkit.TOMLDocument, Dict[str, Any]]:
    """Parse a TOML MCP config file into (document, servers_dict)."""
    if not content.strip():
        return tomlkit.document(), {}
    doc = tomlkit.parse(content)
    servers: Dict[str, Any] = dict(doc.get(top_key, {}))
    return doc, servers


def merge_toml_mcp_servers(
    existing_content: str,
    new_servers: Dict[str, Any],
    top_key: str,
) -> str:
    """Merge new MCP server entries into a TOML config file (always strips type)."""
    doc, _ = parse_toml_mcp(existing_content, top_key)
    if top_key not in doc:
        doc.add(top_key, tomlkit.table())
    mcp: Any = doc[top_key]
    for name, config in _strip_mcp_type(new_servers).items():
        mcp[name] = config
    return tomlkit.dumps(doc)


def strip_rule_frontmatter(content: str) -> str:
    """Strip frontmatter keys that Claude Code rules don't use (e.g. alwaysApply)."""
    fm, body = parse_frontmatter(content)
    if not fm:
        return content
    kept = {k: v for k, v in fm.items() if k in ("name", "description")}
    return render_frontmatter(kept, body) if kept else body


def ensure_cursor_rule_frontmatter(content: str) -> str:
    """Ensure Cursor rule has alwaysApply and description in frontmatter."""
    fm, body = parse_frontmatter(content)
    fm["alwaysApply"] = True
    if "description" not in fm:
        heading = extract_first_heading(body)
        if heading:
            fm["description"] = heading
    return render_frontmatter(fm, body)


def wrap_as_skill(content: str, skill_name: str, always_apply: bool = False) -> str:
    """Wrap content as a Codex SKILL.md with name/description frontmatter."""
    fm, body = parse_frontmatter(content)
    skill_fm: Dict[str, Any] = {}
    skill_fm["name"] = fm.get("name", skill_name)
    desc = fm.get("description") or extract_first_heading(body) or skill_name
    if always_apply:
        desc = "ALWAYS read and follow this skill before acting. " + desc
    skill_fm["description"] = desc
    return render_frontmatter(skill_fm, body)


def safe_write_text(dest: Path, content: str) -> None:
    """Write content to dest atomically via write-then-move.

    Writes to a uniquely-named temp sibling first, then uses os.replace()
    for an atomic rename on the same filesystem.
    """
    tmp = dest.parent / (dest.name + "." + uniq_id(8) + ".tmp")
    try:
        tmp.write_text(content, encoding="utf-8")
        os.replace(tmp, dest)
    except BaseException:
        try:
            tmp.unlink(missing_ok=True)
        except OSError:
            pass
        raise


MIN_REDACT_STARS = 10
"""Minimum number of stars used when redacting a value, hides the length of short secrets."""


def redact_value(val: Any) -> str:
    """Replaces a leaf value with stars.

    Values shorter than MIN_REDACT_STARS always produce that many stars (hides
    length of short secrets). Longer values produce stars matching their length.
    Newlines are preserved.
    """
    s = str(val)
    if "\n" in s:
        return "\n".join(
            ("*" * max(MIN_REDACT_STARS, len(line)) if line else "") for line in s.split("\n")
        )
    return "*" * max(MIN_REDACT_STARS, len(s))


def redact_toml_document(doc: tomlkit.TOMLDocument) -> tomlkit.TOMLDocument:
    """Returns a deep copy of `doc` with all leaf values replaced by stars.

    Configuration placeholder values from TYPE_EXAMPLES (e.g. `<configure me>`)
    are preserved verbatim so the caller can see which fields still need to be set.
    """
    placeholders = set(TYPE_EXAMPLES.values())
    redacted = tomlkit.parse(tomlkit.dumps(doc))

    def _is_placeholder(val: Any) -> bool:
        try:
            return val in placeholders
        except TypeError:
            return False

    def _redact(container: Any) -> None:
        if isinstance(container, dict):
            for key in container:
                val = container[key]
                if isinstance(val, dict):
                    _redact(val)
                elif not _is_placeholder(val):
                    container[key] = redact_value(val)

    _redact(redacted)
    return redacted


def fetch_secrets_list() -> List[TLocationInfo]:
    """Return project-scoped secret file locations, profile-scoped first."""
    locations: List[TLocationInfo] = []
    for info in get_provider_locations():
        if not isinstance(info.provider, SecretsTomlProvider):
            continue
        project_locs = [loc for loc in info.locations if loc.scope == "project"]
        project_locs.sort(key=lambda loc: (loc.profile_name is None, loc.path))
        for loc in project_locs:
            entry = TLocationInfo(
                path=loc.path,
                present=loc.present,
                scope=loc.scope,
            )
            if loc.profile_name is not None:
                entry["profile_name"] = loc.profile_name
            locations.append(entry)
    return locations


def fetch_secrets_view_redacted(path: Optional[str] = None) -> Optional[str]:
    """Return redacted secrets TOML string, or None if not found.

    Without path, returns the unified merged view from SecretsTomlProvider.
    With path, returns that exact file redacted.
    """
    if path:
        try:
            with open(path, "r", encoding="utf-8") as f:
                doc = tomlkit.load(f)
        except FileNotFoundError:
            return None
    else:
        doc = None
        for info in get_provider_locations():
            if not isinstance(info.provider, SecretsTomlProvider):
                continue
            doc = info.provider._config_toml
            break
        if doc is None or len(doc.body) == 0:
            return None
    return tomlkit.dumps(redact_toml_document(doc))


def fetch_secrets_update_fragment(fragment: str, path: str) -> str:
    """Merge a TOML fragment into the secrets file at path.

    Creates the file if needed. Returns the redacted TOML after merge.
    Raises tomlkit.exceptions.TOMLKitError on invalid fragment.
    """
    settings_dir = os.path.dirname(path) or "."
    file_name = os.path.basename(path)
    os.makedirs(settings_dir, exist_ok=True)
    if not os.path.isfile(path):
        with open(path, "w", encoding="utf-8") as f:
            f.write("")
    provider = SettingsTomlProvider(file_name, True, file_name, [settings_dir])
    # allow literal \n (two chars) as newline — agents on Windows/PowerShell
    # can't easily pass real newlines, so we accept the escaped form
    if "\\n" in fragment and "\n" not in fragment:
        fragment = fragment.replace("\\n", "\n")
    parsed = tomlkit.parse(fragment)
    update_dict_nested(provider._config_toml, parsed)
    provider.write_toml()
    return tomlkit.dumps(redact_toml_document(provider._config_toml))


def read_workbench_toolkit_combined_info(toolkit_dir: Path) -> Optional[Dict[str, Any]]:
    """Read plugin.json and toolkit.json from a toolkit directory.

    plugin.json must exist and is the base metadata (validated by Claude).
    toolkit.json is optional and carries dlt-specific fields (`listed`,
    `dependencies`) that are not part of the Claude plugin schema.  When
    present its keys are merged on top of plugin.json.

    Returns None when plugin.json is missing or unreadable.
    """
    plugin_dir = toolkit_dir / ".claude-plugin"
    meta_json = plugin_dir / "plugin.json"
    if not meta_json.exists():
        return None
    try:
        meta: Dict[str, Any] = json.loads(meta_json.read_text(encoding="utf-8"))
    except (ValueError, OSError):
        return None
    # merge dlt-specific toolkit.json when present
    toolkit_json = plugin_dir / "toolkit.json"
    if toolkit_json.is_file():
        extra = json.loads(toolkit_json.read_text(encoding="utf-8"))
        meta.update(extra)
    return meta


def read_workbench_toolkit_mcp_servers(toolkit_dir: Path) -> Dict[str, Any]:
    """Read MCP server definitions from plugin.json mcpServers or standalone .mcp.json/mcp.json."""
    meta = read_workbench_toolkit_combined_info(toolkit_dir)
    if meta and "mcpServers" in meta:
        servers: Dict[str, Any] = meta["mcpServers"]
        return servers
    # check standalone files — .mcp.json (Claude format) then mcp.json
    for name in (".mcp.json", "mcp.json"):
        mcp_file = toolkit_dir / name
        if mcp_file.is_file():
            data = json.loads(mcp_file.read_text(encoding="utf-8"))
            # .mcp.json uses {"mcpServers": {...}} wrapper
            if "mcpServers" in data:
                return data["mcpServers"]  # type: ignore[no-any-return]
            return data  # type: ignore[no-any-return]
    return {}


def build_toolkits_dependency_map(
    toolkits: Dict[str, TToolkitInfo],
) -> Dict[str, List[str]]:
    """Build a map of toolkit name -> list of dependency toolkit names."""
    return {name: list(info.get("dependencies", [])) for name, info in toolkits.items()}


def resolve_toolkit_dependencies(name: str, dep_map: Dict[str, List[str]]) -> List[str]:
    """Return install-order list of dependencies for a toolkit.

    Args:
        name: Toolkit to resolve dependencies for.
        dep_map: Mapping of toolkit name to its direct dependencies.

    Returns:
        Topologically sorted dependency names, excluding the toolkit itself.

    Raises:
        ValueError: On circular dependencies.
    """
    order: List[str] = []
    visited: Set[str] = set()
    path: Set[str] = set()

    def _visit(n: str) -> None:
        if n in path:
            raise ValueError("Circular dependency: %s" % " -> ".join([*path, n]))
        if n in visited:
            return
        path.add(n)
        for dep in dep_map.get(n, []):
            _visit(dep)
        path.discard(n)
        visited.add(n)
        order.append(n)

    _visit(name)
    # remove the toolkit itself — caller installs it separately
    order.remove(name)
    return order


def fetch_workbench_base(location: str, branch: Optional[str]) -> Path:
    """Fetch AI workbench repo and return base Path.

    Thread-safe: git operations on the shared repo directory are serialized.

    Raises:
        FileNotFoundError: When the workbench directory is missing from the repo.
    """
    from dlt.common.libs import git

    branch = branch or DEFAULT_AI_WORKBENCH_BRANCH
    with _workbench_lock:
        src_storage = git.get_fresh_repo_files(location, get_dlt_repos_dir(), branch=branch)
    if not src_storage.has_folder(AI_WORKBENCH_BASE_DIR):
        raise FileNotFoundError(
            "Workbench directory '%s' not found in repo %s" % (AI_WORKBENCH_BASE_DIR, location)
        )
    return Path(src_storage.make_full_path(AI_WORKBENCH_BASE_DIR))


def fetch_workbench_toolkits(
    base: Path, listed_only: bool = False, strict: bool = False
) -> Tuple[Dict[str, TToolkitInfo], List[str]]:
    """Scan workbench directory and return mapping of toolkit name -> TToolkitInfo.

    Args:
        base: Root directory of the workbench repo.
        listed_only: When True, skip toolkits with `listed: false` in metadata.
        strict: When True, raise on invalid toolkit metadata instead of warning.

    Returns:
        Tuple of (toolkits mapping, warnings). Warnings contain messages for
        toolkits with invalid metadata that were skipped.

    Raises:
        ValueError: When `strict` is True and a toolkit has invalid metadata.
    """
    result: Dict[str, TToolkitInfo] = {}
    warnings: List[str] = []
    for entry in sorted(base.iterdir()):
        if not entry.is_dir():
            continue
        meta = read_workbench_toolkit_combined_info(entry)
        if meta is None:
            continue
        if listed_only and meta.get("listed", True) is False:
            continue
        try:
            info = extract_toolkit_info(meta, entry.name)
        except ValueError as ex:
            if strict:
                raise
            warnings.append(str(ex))
            continue
        result[info["name"]] = info
    return result, warnings


def extract_toolkit_info(meta: Dict[str, Any], fallback_name: str) -> TToolkitInfo:
    """Build a TToolkitInfo from a raw metadata dict.

    Args:
        meta: Raw metadata dictionary with toolkit fields.
        fallback_name: Name used in error messages when validation fails.

    Returns:
        Validated toolkit info.

    Raises:
        ValueError: When name, description, or version is missing or empty.
    """
    missing = [k for k in ("name", "description", "version") if not meta.get(k)]
    if missing:
        raise ValueError(
            "plugin.json for %s missing required fields: %s" % (fallback_name, ", ".join(missing))
        )
    tk_meta = TToolkitInfo(
        name=meta["name"],
        version=meta["version"],
        description=meta["description"],
        tags=list(meta.get("keywords", [])),
    )
    if deps := meta.get("dependencies"):
        tk_meta["dependencies"] = list(deps)
    if wes := meta.get("workflow_entry_skill"):
        tk_meta["workflow_entry_skill"] = wes
    return tk_meta


def fetch_workbench_toolkit_info(
    name: str, location: str, branch: Optional[str] = None
) -> Optional[TWorkbenchToolkitInfo]:
    """Return detailed toolkit info from the workbench repo.

    Args:
        name: Toolkit name to look up.
        location: Workbench git repo URL or local path.
        branch: Git branch to fetch. Uses default workbench branch when None.

    Returns:
        Toolkit info with skills, commands, and rules, or None if not found.
    """
    base = fetch_workbench_base(location, branch)
    toolkit_dir = base / name
    meta = read_workbench_toolkit_combined_info(toolkit_dir)
    if meta is None:
        return None

    tk_meta = extract_toolkit_info(meta, name)

    def _components(md_files: List[Path]) -> List[TWorkbenchComponentInfo]:
        return [
            TWorkbenchComponentInfo(name=n, description=d)
            for n, d in (read_md_name_desc(f) for f in md_files)
        ]

    skills: List[TWorkbenchComponentInfo] = []
    skills_dir = toolkit_dir / "skills"
    if skills_dir.is_dir():
        skills = _components(
            [
                p / "SKILL.md"
                for p in sorted(skills_dir.iterdir())
                if p.is_dir() and (p / "SKILL.md").exists()
            ]
        )

    commands: List[TWorkbenchComponentInfo] = []
    cmd_dir = toolkit_dir / "commands"
    if cmd_dir.is_dir():
        commands = _components(sorted(cmd_dir.glob("*.md")))

    rules: List[TWorkbenchComponentInfo] = []
    rules_dir = toolkit_dir / "rules"
    if rules_dir.is_dir():
        rules = _components(sorted(rules_dir.glob("*.md")))

    servers = read_workbench_toolkit_mcp_servers(toolkit_dir)

    info = TWorkbenchToolkitInfo(
        **tk_meta,
        skills=skills,
        commands=commands,
        rules=rules,
        has_ignore=(toolkit_dir / ".claudeignore").is_file(),
    )
    if servers:
        info["mcp_servers"] = servers
    return info


def _toolkits_index_path() -> str:
    return make_dlt_settings_path(TOOLKITS_INDEX_FILE)


def load_toolkits_index() -> Dict[str, TToolkitIndexEntry]:
    """Load the installed toolkits index from the workspace settings directory."""
    path = _toolkits_index_path()
    if not os.path.isfile(path):
        return {}
    try:
        with open(path, "r", encoding="utf-8") as f:
            data: Dict[str, TToolkitIndexEntry] = yaml.safe_load(f)
        if not data:
            return {}
        # restore name from the YAML key (stripped on write)
        for key, entry in data.items():
            entry["name"] = key
        return data
    except (yaml.YAMLError, OSError):
        return {}


def is_toolkit_installed(name: str) -> bool:
    """Check whether a toolkit is recorded in the index."""
    return name in load_toolkits_index()


def save_toolkit_entry(
    toolkit_meta: TToolkitInfo,
    agent: Optional[str] = None,
    files: Optional[Dict[str, Any]] = None,
    mcp_servers: Optional[List[str]] = None,
) -> None:
    """Record that a toolkit was installed or updated in the toolkits index.

    Args:
        toolkit_meta: Core toolkit metadata to persist.
        agent: Name of the AI agent that triggered the install.
        files: Mapping of installed file paths to their content hashes.
        mcp_servers: Names of MCP servers provided by the toolkit.
    """

    index = load_toolkits_index()
    entry: Dict[str, Any] = dict(toolkit_meta)
    name = entry.pop("name")
    entry["installed_at"] = pendulum.now("UTC").isoformat()
    if agent:
        entry["agent"] = agent
    if files:
        entry["files"] = files
    if mcp_servers:
        entry["mcp_servers"] = mcp_servers
    index[name] = entry  # type: ignore[assignment]
    path = _toolkits_index_path()
    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, "w", encoding="utf-8") as f:
        yaml.dump(index, f, allow_unicode=True, default_flow_style=False, sort_keys=False)


_INIT_TOOLKIT = "init"


def read_agents_md_template(workbench_base: Optional[Path]) -> str:
    """Read AGENTS.md template from init toolkit in workbench.

    Raises `FileNotFoundError` when `workbench_base` is `None`
    or the template file is missing.
    """
    if workbench_base is None:
        raise FileNotFoundError("workbench_base is required for AGENTS.md template")
    tpl_path = workbench_base / _INIT_TOOLKIT / "AGENTS.md"
    return tpl_path.read_text(encoding="utf-8")


def fetch_ai_status(project_root: Path) -> TAiStatusInfo:
    """Collect AI setup status for the current workspace.

    Args:
        project_root: Path to the project root for agent detection.

    Returns:
        Status info including dlt version, detected agent, installed toolkits,
        and any readiness warnings.
    """
    from dlt._workspace.cli.ai.agents import _AIAgent  # circular

    index = load_toolkits_index()
    warnings: List[TAiStatusWarning] = []

    # detect agent
    agent_name: Optional[str] = None
    init_entry = index.get(_INIT_TOOLKIT)
    if isinstance(init_entry, dict):
        agent_name = init_entry.get("agent")
    if not agent_name:
        detected = _AIAgent.detect_all(project_root)
        if detected:
            agent_name = detected[0][0].name

    # workspace initialized?
    initialized = os.path.isfile(make_dlt_settings_path("config.toml"))
    if not initialized:
        warnings.append("not_initialized")

    # init toolkit installed?
    has_init = _INIT_TOOLKIT in index
    if not has_init:
        warnings.append("no_init_toolkit")

    # non-init toolkits
    toolkits = {k: v for k, v in index.items() if k != _INIT_TOOLKIT}
    if not toolkits:
        warnings.append("no_toolkits")

    # check MCP availability
    status = TAiStatusInfo(
        dlt_version=dlt_ver,
        agent_name=agent_name,
        initialized=initialized,
        has_init_toolkit=has_init,
        toolkits=toolkits,
        warnings=warnings,
    )
    try:
        from dlt._workspace.mcp import WorkspaceMCP  # noqa: F401

        WorkspaceMCP("dlt")
    except Exception as ex:
        warnings.append("mcp_unavailable")
        status["mcp_error"] = str(ex)

    return status



from __future__ import annotations

from typing import TYPE_CHECKING

from dlt.common.exceptions import MissingDependencyException

try:
    import marimo
    import mowidgets
except ModuleNotFoundError:
    raise MissingDependencyException(
        "dlt.helpers.marimo",
        ["dlt[workspace]"],
    )

from dlt.helpers.marimo._load_package_viewer import app as load_package_viewer
from dlt.helpers.marimo._schema_viewer import app as schema_viewer
from dlt.helpers.marimo._pipeline_selector import app as pipeline_selector

if TYPE_CHECKING:
    from typing import Any


def render(app: marimo.App, *args: Any, **kwargs: Any) -> mowidgets.MoWidget:
    if not isinstance(app, marimo.App):
        raise ValueError("app must be an instance of marimo.App")

    if app is load_package_viewer:
        return load_package_widget(app, *args, **kwargs)
    elif app is schema_viewer:
        return schema_viewer_widget(app, *args, **kwargs)
    elif app is pipeline_selector:
        return pipeline_selector_widget(app, *args, **kwargs)
    else:
        raise ValueError(
            "Unknown app. Must be one of: load_package_viewer, schema_viewer, pipeline_selector"
        )


def pipeline_selector_widget(app: marimo.App, *args: Any, **kwargs: Any) -> mowidgets.MoWidget:
    return mowidgets.widgetize(
        app,
        data_access=True,
        public_variables=["pipeline_path", "pipeline_name", "pipeline_locations"],
    )


def load_package_widget(
    app: marimo.App,
    pipeline_path: str,
    *args: Any,
    **kwargs: Any,
) -> mowidgets.MoWidget:
    return mowidgets.widgetize(app, data_access=True, inputs={"pipeline_path": pipeline_path})


def schema_viewer_widget(
    app: marimo.App,
    pipeline_name: str,
    *args: Any,
    **kwargs: Any,
) -> mowidgets.MoWidget:
    return mowidgets.widgetize(app, data_access=True, inputs={"pipeline_name": pipeline_name})


__all__ = (
    "render",
    "load_package_viewer",
    "schema_viewer",
    "pipeline_selector",
)



import re
from typing import ClassVar

from dlt.common.normalizers.naming import NamingConvention as BaseNamingConvention
from dlt.common.normalizers.naming.snake_case import NamingConvention as SnakeCaseNamingConvention
from dlt.common.typing import REPattern


class NamingConvention(SnakeCaseNamingConvention):
    """Normalizes identifiers according to Weaviate documentation: https://weaviate.io/developers/weaviate/config-refs/schema#class"""

    @property
    def is_case_sensitive(self) -> bool:
        return True

    RESERVED_PROPERTIES = {"id": "__id", "_id": "___id", "_additional": "__additional"}
    RE_UNDERSCORES: ClassVar[REPattern] = re.compile("([^_])__+")
    _STARTS_DIGIT = re.compile("^[0-9]")
    _STARTS_NON_LETTER = re.compile("^[0-9_]")
    _SPLIT_UNDERSCORE_NON_CAP = re.compile("(_[^A-Z])")

    def normalize_identifier(self, identifier: str) -> str:
        """Normalizes Weaviate property name by removing not allowed characters, replacing them by _ and contracting multiple _ into single one
        and lowercasing the first character.

        """
        identifier = BaseNamingConvention.normalize_identifier(self, identifier)
        if identifier in self.RESERVED_PROPERTIES:
            return self.RESERVED_PROPERTIES[identifier]
        norm_identifier = self._base_normalize(identifier)
        if self._STARTS_DIGIT.match(norm_identifier):
            norm_identifier = "p_" + norm_identifier
        norm_identifier = self._lowercase_property(norm_identifier)
        return self.shorten_identifier(norm_identifier, identifier, self.max_length)

    def normalize_table_identifier(self, identifier: str) -> str:
        """Creates Weaviate class name. Runs property normalization and then creates capitalized case name by splitting on _

        https://weaviate.io/developers/weaviate/configuration/schema-configuration#create-a-class
        """
        identifier = BaseNamingConvention.normalize_identifier(self, identifier)
        norm_identifier = self._base_normalize(identifier)
        # norm_identifier = norm_identifier.strip("_")
        norm_identifier = "".join(
            s[1:2].upper() + s[2:] if s and s[0] == "_" else s
            for s in self._SPLIT_UNDERSCORE_NON_CAP.split(norm_identifier)
        )
        norm_identifier = norm_identifier[0].upper() + norm_identifier[1:]
        if self._STARTS_NON_LETTER.match(norm_identifier):
            norm_identifier = "C" + norm_identifier
        return self.shorten_identifier(norm_identifier, identifier, self.max_length)

    def _lowercase_property(self, identifier: str) -> str:
        # lowercase the first letter to follow Weaviate guidelines on properties
        return identifier[0].lower() + identifier[1:]

    def _base_normalize(self, identifier: str) -> str:
        # all characters that are not letters digits or a few special chars are replaced with underscore
        normalized_ident = identifier.translate(self._TR_REDUCE_ALPHABET)
        normalized_ident = self.RE_NON_ALPHANUMERIC.sub("_", normalized_ident)
        # replace trailing _ with x
        stripped_ident = normalized_ident.rstrip("_")
        strip_count = len(normalized_ident) - len(stripped_ident)
        stripped_ident += "x" * strip_count

        # replace consecutive underscores with single one to prevent name clashes with PATH_SEPARATOR
        return self.RE_UNDERSCORES.sub(r"\1_", stripped_ident)



from typing import Callable, List, Dict, NamedTuple, Sequence, Set, Optional, Type

from dlt.common import logger
from dlt.common.configuration.container import Container
from dlt.common.data_writers import (
    DataWriter,
    create_import_spec,
    resolve_best_writer_spec,
    get_best_writer_spec,
    is_native_writer,
)
from dlt.common.destination.utils import prepare_load_table
from dlt.common.metrics import DataWriterMetrics
from dlt.common.schema.utils import new_table
from dlt.common.typing import TLoaderFileFormat
from dlt.common.utils import chunks
from dlt.common.schema.typing import TStoredSchema, TTableSchema
from dlt.common.storages import (
    NormalizeStorage,
    LoadStorage,
    LoadStorageConfiguration,
    NormalizeStorageConfiguration,
    ParsedLoadJobFileName,
)
from dlt.common.schema import TSchemaUpdate, Schema
from dlt.common.normalizers.json import helpers as normalize_helpers

from dlt.normalize.configuration import NormalizeConfiguration
from dlt.normalize.exceptions import NormalizeJobFailed
from dlt.normalize.items_normalizers import (
    ArrowItemsNormalizer,
    FileImportNormalizer,
    JsonLItemsNormalizer,
    ItemsNormalizer,
    ModelItemsNormalizer,
)


class TWorkerRV(NamedTuple):
    schema_updates: List[TSchemaUpdate]
    file_metrics: List[DataWriterMetrics]
    null_only_columns: Dict[str, Set[str]]


def group_worker_files(files: Sequence[str], no_groups: int) -> List[Sequence[str]]:
    # sort files so the same tables are in the same worker
    files = list(sorted(files))

    chunk_size = max(len(files) // no_groups, 1)
    chunk_files = list(chunks(files, chunk_size))
    # distribute the remainder files to existing groups starting from the end
    remainder_l = len(chunk_files) - no_groups
    l_idx = 0
    while remainder_l > 0:
        idx = 0
        for idx, file in enumerate(reversed(chunk_files.pop())):
            chunk_files[-l_idx - idx - remainder_l].append(file)  # type: ignore
        remainder_l -= 1
        l_idx = idx + 1
    return chunk_files


def w_normalize_files(
    config: NormalizeConfiguration,
    normalize_storage_config: NormalizeStorageConfiguration,
    loader_storage_config: LoadStorageConfiguration,
    stored_schema: TStoredSchema,
    load_id: str,
    extracted_items_files: Sequence[str],
) -> TWorkerRV:
    destination_caps = config.destination_capabilities
    schema_updates: List[TSchemaUpdate] = []
    # normalizers are cached per {table_name}.{item_format}
    item_normalizers: Dict[str, ItemsNormalizer] = {}

    preferred_file_format = (
        destination_caps.preferred_loader_file_format
        or destination_caps.preferred_staging_file_format
    )
    # TODO: capabilities.supported_*_formats can be None, it should have defaults
    supported_file_formats = destination_caps.supported_loader_file_formats or []

    # process all files with data items and write to buffered item storage
    with Container().injectable_context(destination_caps):
        schema = Schema.from_stored_schema(stored_schema)
        normalize_storage = NormalizeStorage(False, normalize_storage_config)
        load_storage = LoadStorage(False, supported_file_formats, loader_storage_config)

        def _get_items_normalizer(
            parsed_file_name: ParsedLoadJobFileName, table_schema: TTableSchema
        ) -> ItemsNormalizer:
            item_format = DataWriter.item_format_from_file_extension(parsed_file_name.file_format)

            table_name = table_schema["name"]
            normalizer_key = f"{table_name}.{item_format}"

            if normalizer_key in item_normalizers:
                return item_normalizers[normalizer_key]

            # TODO: extract code that resolves file_format from preferred to utils

            items_preferred_file_format = preferred_file_format
            items_supported_file_formats = supported_file_formats
            if destination_caps.loader_file_format_selector is not None:
                items_preferred_file_format, items_supported_file_formats = (
                    destination_caps.loader_file_format_selector(
                        preferred_file_format,
                        (
                            supported_file_formats.copy()
                            if isinstance(supported_file_formats, list)
                            else supported_file_formats
                        ),
                        table_schema=table_schema,
                    )
                )

            best_writer_spec = None
            if item_format == "file":
                # if we want to import file, create a spec that may be used only for importing
                best_writer_spec = create_import_spec(
                    parsed_file_name.file_format,  # type: ignore[arg-type]
                    items_supported_file_formats,
                )

            config_loader_file_format: TLoaderFileFormat = None
            if file_format := table_schema.get("file_format"):
                # resource has a file format defined so use it
                if file_format == "preferred":
                    # use destination preferred
                    config_loader_file_format = items_preferred_file_format
                else:
                    # use resource format
                    config_loader_file_format = file_format
                logger.info(
                    f"A file format for table {table_name} was specified to {file_format} in the"
                    f" resource so {config_loader_file_format} format being used."
                )

            if config_loader_file_format and best_writer_spec is None:
                # force file format
                if config_loader_file_format in items_supported_file_formats:
                    # TODO: pass supported_file_formats, when used in pipeline we already checked that
                    # but if normalize is used standalone `supported_loader_file_formats` may be unresolved
                    best_writer_spec = get_best_writer_spec(item_format, config_loader_file_format)
                else:
                    logger.warning(
                        f"The configured value `{config_loader_file_format}` "
                        "for `loader_file_format` is not supported for table "
                        f"`{table_name}` and will be ignored. dlt "
                        "will use a supported format instead."
                    )

            if best_writer_spec is None:
                # find best spec among possible formats taking into account destination preference
                best_writer_spec = resolve_best_writer_spec(
                    item_format, items_supported_file_formats, items_preferred_file_format
                )
                # if best_writer_spec.file_format != preferred_file_format:
                #     logger.warning(
                #         f"For data items yielded as {item_format} jobs in file format"
                #         f" {preferred_file_format} cannot be created."
                #         f" {best_writer_spec.file_format} jobs will be used instead."
                #         " This may decrease the performance."
                #     )
            item_storage = load_storage.create_item_storage(best_writer_spec)
            if not is_native_writer(item_storage.writer_cls):
                logger.warning(
                    f"For data items in `{table_name}` yielded as {item_format} and job file format"
                    f" {best_writer_spec.file_format} native writer could not be found. A"
                    f" {item_storage.writer_cls.__name__} writer is used that internally"
                    f" converts {item_format}. This will degrade performance."
                )

            cls: Type[ItemsNormalizer]
            if item_format == "arrow":
                cls = ArrowItemsNormalizer
            elif item_format == "object":
                cls = JsonLItemsNormalizer
            elif item_format == "model":
                cls = ModelItemsNormalizer
            else:
                cls = FileImportNormalizer
            logger.info(
                f"Created items normalizer {cls.__name__} with writer"
                f" {item_storage.writer_cls.__name__} for item format {item_format} and file"
                f" format {item_storage.writer_spec.file_format}"
            )

            norm = item_normalizers[normalizer_key] = cls(
                item_storage,
                load_storage,
                normalize_storage,
                schema,
                load_id,
                config,
            )
            return norm

        def _gather_metrics_and_close(
            parsed_fn: ParsedLoadJobFileName, in_exception: bool
        ) -> List[DataWriterMetrics]:
            writer_metrics: List[DataWriterMetrics] = []
            try:
                try:
                    for normalizer in item_normalizers.values():
                        normalizer.item_storage.close_writers(load_id, skip_flush=in_exception)
                except Exception:
                    # if we had exception during flushing the writers, close them without flushing
                    if not in_exception:
                        for normalizer in item_normalizers.values():
                            normalizer.item_storage.close_writers(load_id, skip_flush=True)
                    raise
                finally:
                    # always gather metrics
                    for normalizer in item_normalizers.values():
                        norm_metrics = normalizer.item_storage.closed_files(load_id)
                        writer_metrics.extend(norm_metrics)
                    for normalizer in item_normalizers.values():
                        normalizer.item_storage.remove_closed_files(load_id)
            except Exception as exc:
                if in_exception:
                    # swallow exception if we already handle exceptions
                    return writer_metrics
                else:
                    # enclose the exception during the closing in job failed exception
                    job_id = parsed_fn.job_id() if parsed_fn else ""
                    raise NormalizeJobFailed(load_id, job_id, str(exc), writer_metrics)
            return writer_metrics

        parsed_file_name: ParsedLoadJobFileName = None
        try:
            root_tables: Set[str] = set()
            for extracted_items_file in extracted_items_files:
                parsed_file_name = ParsedLoadJobFileName.parse(extracted_items_file)
                # normalize table name in case the normalization changed
                # NOTE: this is the best we can do, until a full lineage information is in the schema
                root_table_name = normalize_helpers.normalize_table_identifier(
                    schema, schema.naming, parsed_file_name.table_name
                )
                root_tables.add(root_table_name)
                root_table = stored_schema["tables"].get(root_table_name) or new_table(
                    root_table_name
                )
                # prepare table
                root_table = prepare_load_table(
                    stored_schema["tables"], root_table, destination_caps
                )
                normalizer = _get_items_normalizer(
                    parsed_file_name,
                    root_table,
                )
                logger.debug(
                    f"Processing extracted items in {extracted_items_file} in load_id"
                    f" {load_id} with table name {root_table_name} and schema {schema.name}"
                )
                partial_updates = normalizer(extracted_items_file, root_table_name)
                schema_updates.extend(partial_updates)
                logger.debug(f"Processed file {extracted_items_file}")
        except Exception as exc:
            job_id = parsed_file_name.job_id() if parsed_file_name else ""
            writer_metrics = _gather_metrics_and_close(parsed_file_name, in_exception=True)
            raise NormalizeJobFailed(load_id, job_id, str(exc), writer_metrics) from exc
        else:
            writer_metrics = _gather_metrics_and_close(parsed_file_name, in_exception=False)
        finally:
            for normalizer in item_normalizers.values():
                normalizer.close()

        # gather null-only columns from all normalizers
        all_null_only: Dict[str, Set[str]] = {}
        for normalizer in item_normalizers.values():
            for table_name, cols in normalizer.null_only_columns.items():
                all_null_only.setdefault(table_name, set()).update(cols)

        logger.info(f"Processed all items in {len(extracted_items_files)} files")
        return TWorkerRV(schema_updates, writer_metrics, all_null_only)



import contextlib
from typing import List
from packaging.requirements import Requirement
from packaging.version import Version

from dlt.common.runners import Venv
from dlt.common.destination.client import DestinationClientDwhConfiguration
from dlt.common.configuration.specs import CredentialsWithDefault
from dlt.common.typing import TSecretStrValue, ConfigValue
from dlt.version import get_installed_requirement_string

from dlt.helpers.dbt.runner import create_runner, DBTPackageRunner

DEFAULT_DBT_VERSION = ">=1.7,<2"

# a map of destination names to dbt package names in case they don't match the pure destination name
DBT_DESTINATION_MAP = {
    "athena": "athena-community",
    "motherduck": "duckdb",
    "mssql": "sqlserver",
}


def _default_profile_name(credentials: DestinationClientDwhConfiguration) -> str:
    profile_name = credentials.destination_type
    # in case of credentials with default add default to the profile name
    if isinstance(credentials.credentials, CredentialsWithDefault):
        if credentials.credentials.has_default_credentials():
            profile_name += "_default"
    elif profile_name == "snowflake":
        if getattr(credentials.credentials, "private_key", None):
            # snowflake with private key is a separate profile
            profile_name += "_pkey"
    return profile_name


def _create_dbt_deps(
    destination_types: List[str], dbt_version: str = DEFAULT_DBT_VERSION
) -> List[str]:
    if dbt_version:
        # if parses as version use "==" operator
        with contextlib.suppress(ValueError):
            Version(dbt_version)
            dbt_version = "==" + dbt_version
    else:
        dbt_version = ""

    # add version only to the core package. the other packages versions must be resolved by pip
    all_packages = ["core" + dbt_version] + destination_types
    for idx, package in enumerate(all_packages):
        package = "dbt-" + DBT_DESTINATION_MAP.get(package, package)
        # verify package
        Requirement(package)
        all_packages[idx] = package

    dlt_requirement = get_installed_requirement_string(allow_earlier=True)
    # get additional requirements
    additional_deps: List[str] = []
    if "duckdb" in destination_types or "motherduck" in destination_types:
        from importlib.metadata import version as pkg_version

        # force locally installed duckdb
        additional_deps = ["duckdb" + "==" + pkg_version("duckdb")]

    return all_packages + additional_deps + [dlt_requirement]


def restore_venv(
    venv_dir: str, destination_types: List[str], dbt_version: str = DEFAULT_DBT_VERSION
) -> Venv:
    venv = Venv.restore(venv_dir)
    venv.add_dependencies(_create_dbt_deps(destination_types, dbt_version))
    return venv


def create_venv(
    venv_dir: str, destination_types: List[str], dbt_version: str = DEFAULT_DBT_VERSION
) -> Venv:
    return Venv.create(venv_dir, _create_dbt_deps(destination_types, dbt_version))


def package_runner(
    venv: Venv,
    destination_configuration: DestinationClientDwhConfiguration,
    working_dir: str,
    package_location: str,
    package_repository_branch: str = ConfigValue,
    package_repository_ssh_key: TSecretStrValue = "",
    auto_full_refresh_when_out_of_sync: bool = ConfigValue,
) -> DBTPackageRunner:
    default_profile_name = _default_profile_name(destination_configuration)
    return create_runner(
        venv,
        destination_configuration,
        working_dir,
        package_location,
        package_repository_branch=package_repository_branch,
        package_repository_ssh_key=package_repository_ssh_key,
        package_profile_name=default_profile_name,
        auto_full_refresh_when_out_of_sync=auto_full_refresh_when_out_of_sync,
    )



from typing import Protocol, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from dlt.pipeline.trace import PipelineTrace, PipelineStepTrace
    from dlt.pipeline.typing import TPipelineStep
    from dlt.common.pipeline import SupportsPipeline
else:
    PipelineTrace = PipelineStepTrace = TPipelineStep = SupportsPipeline = Any


class SupportsTracking(Protocol):
    def on_start_trace(
        self, trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
    ) -> None: ...

    def on_start_trace_step(
        self, trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
    ) -> None: ...

    def on_end_trace_step(
        self,
        trace: PipelineTrace,
        step: PipelineStepTrace,
        pipeline: SupportsPipeline,
        step_info: Any,
        send_state: bool,
    ) -> None: ...

    def on_end_trace(
        self, trace: PipelineTrace, pipeline: SupportsPipeline, send_state: bool
    ) -> None: ...



"""SQL client for Fabric Warehouse - extends Synapse SQL client"""

from typing import TYPE_CHECKING

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.destinations.impl.synapse.sql_client import SynapseSqlClient

if TYPE_CHECKING:
    from dlt.destinations.impl.fabric.configuration import FabricCredentials


class FabricSqlClient(SynapseSqlClient):
    """SQL client for Microsoft Fabric Warehouse

    Inherits all behavior from Synapse since Fabric Warehouse is built on Synapse technology.
    """

    def __init__(
        self,
        dataset_name: str,
        staging_dataset_name: str,
        credentials: "FabricCredentials",
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        # FabricCredentials has all required attributes: database, to_odbc_dsn(), connect_timeout
        super().__init__(dataset_name, staging_dataset_name, credentials, capabilities)  # type: ignore[arg-type]
        self.credentials: "FabricCredentials" = credentials  # type: ignore[assignment]



"""The Debug Pipeline Template will load a column with each datatype to your destination."""

# mypy: disable-error-code="no-untyped-def,arg-type"

import dlt

from dlt.common import Decimal


@dlt.resource(write_disposition="append", name="all_datatypes")
def resource():
    """this is the test data for loading validation, delete it once you yield actual data"""
    yield [
        {
            "col1": 989127831,
            "col2": 898912.821982,
            "col3": True,
            "col4": "2022-05-23T13:26:45.176451+00:00",
            "col5": "string data \n \r  🦆",
            "col6": Decimal("2323.34"),
            "col7": b"binary data \n \r ",
            "col8": 2**56 + 92093890840,
            "col9": {
                "json": [1, 2, 3, "a"],
                "link": (
                    "?commen\ntU\nrn=urn%3Ali%3Acomment%3A%28acti\012 \6"
                    " \\vity%3A69'08444473\n\n551163392%2C6n \r 9085"
                ),
            },
            "col10": "2023-02-27",
            "col11": "13:26:45.176451",
        }
    ]


@dlt.source
def source():
    """A source function groups all resources into one schema."""
    return resource()


def load_all_datatypes() -> None:
    # specify the pipeline name, destination and dataset name when configuring pipeline,
    # otherwise the defaults will be used that are derived from the current script name
    pipeline = dlt.pipeline(
        pipeline_name="debug",
        destination="duckdb",
        dataset_name="debug_data",
    )

    data = list(source().all_datatypes)

    # print the data yielded from resource without loading it
    print(data)  # noqa: T201

    # run the pipeline with your parameters
    load_info = pipeline.run(source())

    # pretty print the information on data that was loaded
    print(load_info)  # noqa: T201


if __name__ == "__main__":
    load_all_datatypes()



import pathlib
from collections.abc import Mapping
from typing import Any, Literal, Union, overload, Optional, cast

import dlt
from dlt.common.exceptions import MissingDependencyException
from dlt.common.schema.typing import (
    C_DLT_LOAD_ID,
    VERSION_TABLE_NAME,
    TColumnSchema,
    TTableReference,
    TTableSchema,
    TStoredSchema,
)
from dlt.common.schema.utils import (
    create_load_table_reference,
    create_version_and_loads_hash_reference,
    create_version_and_loads_schema_name_reference,
    get_data_and_dlt_tables,
    get_first_column_name_with_prop,
    group_tables_by_resource,
    create_root_child_reference,
    create_parent_child_reference,
    is_complete_column,
    is_nested_table,
    remove_column_defaults,
)
from dlt.common.utils import without_none

try:
    from pydbml import Database  # type: ignore[import-untyped]
    from pydbml.classes import Table, Column, Reference, TableGroup  # type: ignore[import-untyped]
except ModuleNotFoundError:
    raise MissingDependencyException("PyDBML", ["pydbml>=1.2.0"])


__all__ = (
    "schema_to_dbml",
    "export_to_dbml",
)

# dlt sets no `data_type` field for columns that haven't seen data yet.
# Since DBML doesn't support empty `type` field, we set a constant
# NOTE should we use `UNKNOWN` or `None` or something else?
UNKNOWN_DATA_TYPE = "UNKNOWN"
TDBMLReferenceCardinality = Literal["<", ">", "-", "<>"]


def _stringify_dict(d: Mapping[str, Any]) -> dict[str, str]:
    return {k: str(v) for k, v in d.items()}


def _destringify_dict(d: Mapping[str, str]) -> dict[str, Union[bool, None, str]]:
    output = {}
    for k, v in d.items():
        value: Union[bool, None, str]
        if v == "True":
            value = True
        elif v == "False":
            value = False
        elif v == "None":
            value = None
        else:
            value = str(v)

        output[k] = value

    return output


def _to_dbml_column(column_hints: TColumnSchema) -> Column:
    """Convert a dlt column schema to a DBML column

    `properties` include remaining dlt hints. They can be rendered by `PyDBML`, but
    most DBML frontends will raised errors when included.
    """
    # remove hints that can be retrieved from DBML `Column()`
    properties = remove_column_defaults(
        cast(
            TColumnSchema,
            {
                k: v
                for k, v in column_hints.items()
                if k
                not in (
                    "name",
                    "data_type",
                    "unique",
                    "nullable",
                    "primary_key",
                    "description",
                    "x-normalizer",
                    "x-loader",
                    "x-extractor",
                )
            },
        )
    )
    return Column(
        name=column_hints["name"],
        type=column_hints.get("data_type", UNKNOWN_DATA_TYPE),
        unique=bool(column_hints.get("unique")),
        not_null=not bool(column_hints.get("nullable", True)),
        pk=bool(column_hints.get("primary_key")),
        note=column_hints.get("description"),
        properties=None if properties == {} else _stringify_dict(properties),
    )


def _from_dbml_column(column: Column) -> TColumnSchema:
    """Convert a DBML column to a dlt column schema"""
    hints = cast(
        TColumnSchema,
        {
            "name": column.name,
            "data_type": None if column.type == UNKNOWN_DATA_TYPE else column.type,
            "unique": column.unique,
            "nullable": not column.not_null,
            "primary_key": column.pk,
            "description": None if column.note.text == "" else column.note.text,
            # `column.properties` should contain keys mutually exclusive with the above
            **_destringify_dict(column.properties),
        },
    )
    return cast(TColumnSchema, remove_column_defaults(hints))


def _to_dbml_table(table_hints: TTableSchema) -> Table:
    """Convert a dlt table schema to a DBML table

    `properties` include remaining dlt hints. They can be rendered by `PyDBML`, but
    most DBML frontends will raised errors when included.
    """
    # remove hints that can be retrieved from DBML `Table()`
    properties = {
        k: v for k, v in table_hints.items() if k not in ("name", "columns", "description")
    }
    return Table(
        name=table_hints["name"],
        columns=[
            _to_dbml_column(col)
            for col in table_hints["columns"].values()
            if is_complete_column(col)
        ],
        note=table_hints.get("description"),
        properties=None if properties == {} else _stringify_dict(properties),
    )


def _from_dbml_table(table: Table) -> TTableSchema:
    """Convert a DBML table to a dlt table schema"""
    original_hints = _destringify_dict(table.properties)
    original_hints.pop("columns", None)
    columns = {col.name: _from_dbml_column(col) for col in table.columns}
    retrieved_hints = {
        "name": table.name,
        "columns": columns,
        "description": None if table.note.text == "" else table.note.text,
        **original_hints,
    }
    return cast(TTableSchema, without_none(retrieved_hints))


# TODO store cardinality directly on TTTableReference
# NOTE iterating over all tables each time we create a reference can become inefficient
def _to_dbml_reference(
    *,
    from_table_name: str,
    reference: TTableReference,
    tables: list[Table],
    cardinality: TDBMLReferenceCardinality,
) -> Reference:
    """Convert a dlt table reference to a DBML reference

    NOTE DBML can represent relationship cardinality, but `dlt` doesn't store this information.
    The calling function specifies the loosest possible cardinality.
    """
    from_columns = reference["columns"]
    to_table = reference["referenced_table"]
    to_columns = reference["referenced_columns"]

    from_dbml_columns: list[Column] = []
    to_dbml_columns: list[Column] = []

    for table in tables:
        if table.name == from_table_name:
            for column in table.columns:
                if column.name == from_columns[0]:
                    from_dbml_columns.append(column)

        if table.name == to_table:
            for column in table.columns:
                if column.name == to_columns[0]:
                    to_dbml_columns.append(column)

    if not from_dbml_columns:
        raise ValueError(f"Column `{from_columns}` not found in table `{from_table_name}`")

    if not to_dbml_columns:
        raise ValueError(f"Column `{to_columns}` not found in table `{to_table}`")

    return Reference(type=cardinality, col1=from_dbml_columns, col2=to_dbml_columns)


def _from_dbml_reference(reference: Reference) -> TTableReference:
    """Convert a DBML reference to a dlt table reference"""
    return TTableReference(
        table=reference.table1.name,
        referenced_table=reference.table2.name,
        columns=[col.name for col in reference.col1],
        referenced_columns=[col.name for col in reference.col2],
    )


# NOTE `TableGroup` seem to not be displayed on `dbdiagram.io`
def _group_tables_by_resource(schema: TStoredSchema, db: Database) -> list[TableGroup]:
    """Create DBML table groups for dlt tables.

    Data tables are grouped by resource.
    Internal dlt tables are grouped together despite each having its own resource.
    """
    _, dlt_tables = get_data_and_dlt_tables(schema["tables"])
    dlt_table_names = [t["name"] for t in dlt_tables]
    table_groups = []

    # data tables groups
    for resource, tables in group_tables_by_resource(schema["tables"]).items():
        group_members_table_name = [table["name"] for table in tables]
        group_members_dbml_tables: list[Table] = []
        for dbml_table in db.tables:
            if dbml_table.name in dlt_table_names:
                continue

            if dbml_table.name in group_members_table_name:
                group_members_dbml_tables.append(dbml_table)

        # avoid creating empty groups
        if group_members_dbml_tables:
            table_groups.append(TableGroup(name=resource, items=group_members_dbml_tables))

    # internal dlt tables group
    dlt_dbml_tables = [dbml_table for dbml_table in db.tables if dbml_table.name in dlt_table_names]
    dlt_group = TableGroup(name="_dlt", items=dlt_dbml_tables)
    table_groups.append(dlt_group)

    return table_groups


def _add_tables(
    schema: TStoredSchema,
    dbml_schema: Database,
    *,
    include_dlt_tables: bool,
) -> Database:
    """Add a DBML Table for each dlt table.

    The operation is in-place and returns the `dbml_schema` object passed as input.

    Args:
        schema: dlt schema to convert
        dbml_schema: DBML schema
        include_dlt_tables: If True, include data tables and internal dlt tables.

    Returns:
        DBML schema
    """
    data_tables, dlt_tables = get_data_and_dlt_tables(schema["tables"])

    if include_dlt_tables is True:
        # order allows to keep _dlt tables at the end
        tables = data_tables + dlt_tables
    else:
        tables = data_tables

    for table in tables:
        # skip tables with no complete columns
        if not any(is_complete_column(c) for c in table.get("columns", {}).values()):
            continue

        dbml_table = _to_dbml_table(table)
        dbml_schema.add_table(dbml_table)

    return dbml_schema


def _add_references(
    schema: TStoredSchema,
    dbml_schema: Database,
    *,
    include_internal_dlt_ref: bool,
    include_parent_child_ref: bool,
    include_root_child_ref: bool,
) -> Database:
    """Add a DBML Reference for each dlt reference.

    The operation is in-place and returns the `dbml_schema` object passed as input.

    Args:
        schema: dlt schema to convert
        dbml_schema: DBML schema to add references to
        include_internal_dlt_ref: If True, include references from `root._dlt_load_id` to `_dlt_loads.load_id`
            If `_dlt_version` and `_dlt_loads` are present, include references between them.
        include_parent_child_ref: If True, include references from `child._dlt_parent_id` to `parent._dlt_id`
        include_root_child_ref: If True, include references from `child._dlt_root_id` to `root._dlt_id`

    Returns:
        DBML schema
    """
    # need to create all tables first because `Reference` references `Table` objects
    for table in schema["tables"].values():
        table_name: str = table["name"]
        for reference in table.get("references", []):
            # user-defined references can have arbitrary cardinality and may incorrectly describe the data
            dbml_reference = _to_dbml_reference(
                tables=dbml_schema.tables,
                from_table_name=table_name,
                reference=reference,
                cardinality="<>",  # because it's the loosest
            )
            dbml_schema.add_reference(dbml_reference)

        # link root -> loads table
        if include_internal_dlt_ref is True and bool(table["columns"].get(C_DLT_LOAD_ID)):
            # root table contains 1 to many rows associated with a single row in loads table
            # possible cardinality: `-` (1-to-1) or `>` (m-to-1)
            dbml_reference = _to_dbml_reference(
                tables=dbml_schema.tables,
                from_table_name=table_name,
                reference=create_load_table_reference(table),
                cardinality=">",  # m-to-1
            )
            dbml_schema.add_reference(dbml_reference)

        # link child -> parent
        if include_parent_child_ref is True and is_nested_table(table):
            # child table contains 1 to many rows associated with has a single row in parent table
            # possible cardinality: `-` (1-to-1) or `>` (m-to-1)
            dbml_parent_reference = _to_dbml_reference(
                tables=dbml_schema.tables,
                from_table_name=table_name,
                reference=create_parent_child_reference(schema["tables"], table_name),
                cardinality=">",  # m-to-1
            )
            dbml_schema.add_reference(dbml_parent_reference)

        # link child -> root
        if (
            include_root_child_ref is True
            and is_nested_table(table)
            # the table must have a root key column; can be enabled via `@dlt.source(root_key=True)` or write_disposition
            and get_first_column_name_with_prop(table, "root_key")
        ):
            # child table contains 1 to many rows associated with has a single row in root table
            # possible cardinality: `-` (1-to-1) or `>` (m-to-1)
            dbml_root_reference = _to_dbml_reference(
                tables=dbml_schema.tables,
                from_table_name=table_name,
                reference=create_root_child_reference(schema["tables"], table_name),
                cardinality=">",  # m-to-1
            )
            dbml_schema.add_reference(dbml_root_reference)

    # generate links between internal dlt tables
    if (
        VERSION_TABLE_NAME in schema["tables"]
        and VERSION_TABLE_NAME in schema["tables"]
        and include_internal_dlt_ref is True
    ):
        # a schema version hash can have multiple runs in the loads table
        # schema version hash is unique
        # possible: cardinality: `-` (1-to-1) or `<` (1-to-m)
        dbml_version_and_loads_hash_ref = _to_dbml_reference(
            tables=dbml_schema.tables,
            from_table_name=VERSION_TABLE_NAME,
            reference=create_version_and_loads_hash_reference(schema["tables"]),
            cardinality="<",
        )
        # a schema name can have multiple multiple runs in the loads table
        # schema name is not unique; it can have multiple version hash
        # possible: cardinality: `-` (1-to-1), `<` (1-to-m), or `<>` (m-to-n)
        dbml_version_and_loads_schema_name_ref = _to_dbml_reference(
            tables=dbml_schema.tables,
            from_table_name=VERSION_TABLE_NAME,
            reference=create_version_and_loads_schema_name_reference(schema["tables"]),
            cardinality="<>",
        )

        dbml_schema.add_reference(dbml_version_and_loads_hash_ref)
        dbml_schema.add_reference(dbml_version_and_loads_schema_name_ref)

    return dbml_schema


def _add_table_groups(schema: TStoredSchema, dbml_schema: Database) -> Database:
    """Add a DBML TableGroup to group tables by resources

    The operation is in-place and returns the `dbml_schema` object passed as input.

    Args:
        schema: dlt schema to convert
        dbml_schema: DBML schema

    Returns:
        DBML schema
    """
    for table_group in _group_tables_by_resource(schema, dbml_schema):
        # operation is inplace
        dbml_schema.add_table_group(table_group)

    return dbml_schema


def schema_to_dbml(
    schema: TStoredSchema,
    *,
    allow_custom_dbml_properties: bool = False,
    include_dlt_tables: bool = True,
    include_internal_dlt_ref: bool = True,
    include_parent_child_ref: bool = True,
    include_root_child_ref: bool = True,
    group_by_resource: bool = False,
) -> Database:
    """Convert a dlt.Schema object to a PyDBML representation.

    Args:
        schema: dlt schema to convert
        allow_custom_dbml_properties: If True, store all dlt metadata on DBML properties. This will
            cause rendering errors in most DBML frontends. If True, dlt to DBML all metadata is preserved.
            If False, the conversion loses some details to reconstruct the dlt schema.
        include_dlt_tables: If True, include data tables and internal dlt tables. This will influence table
            references and groups produced.
        include_internal_dlt_ref: If True, include references between tables `_dlt_version`, `_dlt_loads` and `_dlt_pipeline_state`
        include_parent_child_ref: If True, include references from `child._dlt_parent_id` to `parent._dlt_id`
        include_root_child_ref: If True, include references from `child._dlt_root_id` to `root._dlt_id`
        group_by_resource: If True, group tables by resource.

    Returns:
        DBML schema
    """

    dbml_schema = Database(allow_properties=allow_custom_dbml_properties)

    _add_tables(schema, dbml_schema, include_dlt_tables=include_dlt_tables)
    _add_references(
        schema,
        dbml_schema,
        include_internal_dlt_ref=include_internal_dlt_ref,
        include_parent_child_ref=include_parent_child_ref,
        include_root_child_ref=include_root_child_ref,
    )
    if group_by_resource is True:
        _add_table_groups(schema, dbml_schema)

    return dbml_schema


# TODO convert DBML back to dlt schema
# def dbml_to_schema(dbml_schema: Union[Database, str]) -> TStoredSchema:
#     return


@overload
def export_to_dbml(schema: dlt.Schema, *, path: None = None) -> str:
    """Convert `dlt.Schema` to DBML and return the string"""
    ...


@overload
def export_to_dbml(schema: dlt.Schema, *, path: Union[pathlib.Path, str]) -> str:
    """Convert `dlt.Schema` to DBML and write result to path at specified `path`"""
    ...


# TODO this function might not be needed
def export_to_dbml(
    schema: dlt.Schema,
    *,
    path: Optional[Union[pathlib.Path, str]] = None,
    allow_custom_dbml_properties: bool = False,
    include_dlt_tables: bool = True,
    include_internal_dlt_ref: bool = True,
    include_parent_child_ref: bool = True,
    include_root_child_ref: bool = True,
    group_by_resource: bool = False,
) -> str:
    """Convert a `dlt.Schema` to a a DBML string and return its value.

    Args:
        schema: dlt schema to convert
        path: If specified, write the DBML string to the file at the specified path.
        allow_custom_dbml_properties: If True, store all dlt metadata on DBML properties. This will
            cause rendering errors in most DBML frontends. If True, dlt to DBML all metadata is preserved.
            If False, the conversion loses some details to reconstruct the dlt schema.
        include_dlt_tables: If True, include data tables and internal dlt tables. This will influence table
            references and groups produced.
        include_internal_dlt_ref: If True, include references between tables `_dlt_version`, `_dlt_loads` and `_dlt_pipeline_state`
        include_parent_child_ref: If True, include references from `child._dlt_parent_id` to `parent._dlt_id`
        include_root_child_ref: If True, include references from `child._dlt_root_id` to `root._dlt_id`
        group_by_resource: If True, group tables by resource.

    Returns:
        DBML string
    """
    stored_schema = schema.to_dict()
    dbml_schema = schema_to_dbml(
        stored_schema,
        allow_custom_dbml_properties=allow_custom_dbml_properties,
        include_dlt_tables=include_dlt_tables,
        include_internal_dlt_ref=include_internal_dlt_ref,
        include_parent_child_ref=include_parent_child_ref,
        include_root_child_ref=include_root_child_ref,
        group_by_resource=group_by_resource,
    )
    dbml_string: str = dbml_schema.dbml

    if path is None:
        return dbml_string

    path = pathlib.Path(path)
    if not path.exists():
        path.parent.mkdir(parents=True, exist_ok=True)

    if path.suffix != ".dbml":
        path = path.with_suffix(".dbml")

    path.write_text(dbml_string)
    return dbml_string



from __future__ import annotations

from copy import copy, deepcopy
from typing import (
    Callable,
    ClassVar,
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Any,
    cast,
)

from dlt.common.schema.migrations import migrate_schema
from dlt.common.utils import extend_list_deduplicated, simple_repr, without_none
from dlt.common.typing import (
    DictStrAny,
    StrAny,
    REPattern,
    TDataItem,
    get_type_hints,
)
from dlt.common.normalizers import TNormalizersConfig, NamingConvention
from dlt.common.normalizers.json import DataItemNormalizer, TNormalizedRowIterator
from dlt.common.schema import utils
from dlt.common.data_types import TDataType
from dlt.common.schema.utils import is_dlt_table_or_column
from dlt.common.data_types import TDataType
from dlt.common.schema.typing import (
    DLT_NAME_PREFIX,
    SCHEMA_ENGINE_VERSION,
    LOADS_TABLE_NAME,
    VERSION_TABLE_NAME,
    PIPELINE_STATE_TABLE_NAME,
    TPartialTableSchema,
    TSchemaContractEntities,
    TSchemaEvolutionMode,
    TSchemaSettings,
    TSimpleRegex,
    TTableReferenceStandalone,
    TStoredSchema,
    TSchemaTables,
    TTableReference,
    TTableSchema,
    TTableSchemaColumns,
    TColumnSchema,
    TColumnProp,
    TColumnDefaultHint,
    TTypeDetections,
    TSchemaContractDict,
    TSchemaContract,
    TColumnType,
)
from dlt.common.schema.exceptions import (
    IncompatibleSchemaException,
    InvalidSchemaName,
    ParentTableNotFoundException,
    SchemaCorruptedException,
    TableIdentifiersFrozen,
    TableNotFound,
)
from dlt.common.schema.utils import is_complete_column
from dlt.common.schema.normalizers import import_normalizers, configured_normalizers
from dlt.common.schema.exceptions import DataValidationError
from dlt.common.validation import validate_dict


DEFAULT_SCHEMA_CONTRACT_MODE: TSchemaContractDict = {
    "tables": "evolve",
    "columns": "evolve",
    "data_type": "evolve",
}


class Schema:
    ENGINE_VERSION: ClassVar[int] = SCHEMA_ENGINE_VERSION

    naming: NamingConvention
    """Naming convention used by the schema to normalize identifiers"""
    data_item_normalizer: DataItemNormalizer[Any]
    """Data item normalizer used by the schema to create tables"""

    version_table_name: str
    """Normalized name of the version table"""
    loads_table_name: str
    """Normalized name of the loads table"""
    state_table_name: str
    """Normalized name of the dlt state table"""

    _schema_name: str
    _dlt_tables_prefix: str
    _stored_version: int  # version at load time
    _stored_version_hash: str  # version hash at load time
    _stored_previous_hashes: Optional[List[str]]  # list of ancestor hashes of the schema
    _imported_version_hash: str  # version hash of recently imported schema
    _schema_description: str  # optional schema description
    _schema_tables: TSchemaTables
    _settings: (
        TSchemaSettings  # schema settings to hold default hints, preferred types and other settings
    )

    # list of preferred types: map regex on columns into types
    _compiled_preferred_types: List[Tuple[REPattern, TDataType]]
    # compiled default hints
    _compiled_hints: Dict[TColumnDefaultHint, Sequence[REPattern]]
    # compiled exclude filters per table
    _compiled_excludes: Dict[str, Sequence[REPattern]]
    # compiled include filters per table
    _compiled_includes: Dict[str, Sequence[REPattern]]
    # type detections
    _type_detections: Sequence[TTypeDetections]

    # normalizers config
    _normalizers_config: TNormalizersConfig

    def __init__(self, name: str, normalizers: TNormalizersConfig = None) -> None:
        self._reset_schema(name, normalizers)

    @classmethod
    def from_dict(
        cls,
        d: DictStrAny,
        remove_processing_hints: bool = False,
        bump_version: bool = True,
        validate_schema: bool = True,
    ) -> "Schema":
        # upgrade engine if needed
        stored_schema = migrate_schema(d, d["engine_version"], cls.ENGINE_VERSION)
        # verify schema
        if validate_schema:
            utils.validate_stored_schema(stored_schema)
        # add defaults
        stored_schema = utils.apply_defaults(stored_schema)
        # remove processing hints that could be created by normalize and load steps
        if remove_processing_hints:
            utils.remove_processing_hints(stored_schema["tables"])

        # bump version if modified
        if bump_version:
            utils.bump_version_if_modified(stored_schema)
        return cls.from_stored_schema(stored_schema)

    @classmethod
    def from_stored_schema(cls, stored_schema: TStoredSchema) -> "Schema":
        # create new instance from dict
        self: Schema = cls(
            stored_schema["name"], normalizers=stored_schema.get("normalizers", None)
        )
        self._from_stored_schema(stored_schema)
        return self

    def replace_schema_content(
        self, schema: "Schema", link_to_replaced_schema: bool = False
    ) -> None:
        """Replaces content of the current schema with `schema` content. Does not compute new schema hash and
        does not increase the numeric version. Optionally will link the replaced schema to incoming schema
        by keeping its hash in prev hashes and setting stored hash to replaced schema hash.
        """
        # do not bump version so hash from `schema` is preserved
        stored_schema = schema.to_dict(bump_version=False)
        if link_to_replaced_schema:
            replaced_version_hash = self.version_hash
            # do not store hash if the replaced schema is identical
            if schema.version_hash != replaced_version_hash:
                utils.store_prev_hash(stored_schema, replaced_version_hash)
        self._reset_schema(schema.name, schema._normalizers_config)
        self._from_stored_schema(stored_schema)

    def normalize_data_item(
        self, item: TDataItem, load_id: str, table_name: str
    ) -> TNormalizedRowIterator:
        return self.data_item_normalizer.normalize_data_item(item, load_id, table_name)

    def apply_schema_contract(
        self,
        schema_contract: TSchemaContractDict,
        partial_table: TPartialTableSchema,
        data_item: TDataItem = None,
        raise_on_freeze: bool = True,
    ) -> Tuple[
        TPartialTableSchema, List[Tuple[TSchemaContractEntities, str, TSchemaEvolutionMode]]
    ]:
        """
        Checks if `schema_contract` allows for the `partial_table` to update the schema. It applies the contract dropping
        the affected columns or the whole `partial_table`. It generates and returns a set of filters that should be applied to incoming data in order to modify it
        so it conforms to the contract. `data_item` is provided only as evidence in case DataValidationError is raised.

        Example `schema_contract`:
        {
            "tables": "freeze",
            "columns": "evolve",
            "data_type": "discard_row"
        }

        Settings for table affects new tables, settings for column affects new columns and settings for data_type affects new variant columns. Each setting can be set to one of:
        * evolve: allow all changes
        * freeze: allow no change and fail the load
        * discard_row: allow no schema change and filter out the row
        * discard_value: allow no schema change and filter out the value but load the rest of the row

        Returns a tuple where a first element is modified partial table and the second is a list of filters. The modified partial may be None in case the
        whole table is not allowed.
        Each filter is a tuple of (table|columns, entity name, freeze | discard_row | discard_value).
        Note: by default `freeze` immediately raises DataValidationError which is convenient in most use cases

        """
        # default settings allow all evolutions, skip all else
        if schema_contract == DEFAULT_SCHEMA_CONTRACT_MODE:
            return partial_table, []

        assert partial_table
        table_name = partial_table["name"]
        existing_table: TTableSchema = self._schema_tables.get(table_name, None)

        # table is new when not yet exist or
        is_new_table = not existing_table or self.is_new_table(table_name)
        # check case where we have a new table
        if is_new_table and schema_contract["tables"] != "evolve":
            if raise_on_freeze and schema_contract["tables"] == "freeze":
                raise DataValidationError(
                    self.name,
                    table_name,
                    None,
                    "tables",
                    "freeze",
                    None,
                    schema_contract,
                    data_item,
                    f"Can't add table `{table_name}` because `tables` are frozen.",
                )
            # filter tables with name below
            return None, [("tables", table_name, schema_contract["tables"])]

        column_mode, data_mode = schema_contract["columns"], schema_contract["data_type"]
        # allow to add new columns when table is new or if columns are allowed to evolve once
        if is_new_table or existing_table.get("x-normalizer", {}).get("evolve-columns-once", False):
            column_mode = "evolve"

        # check if we should filter any columns,
        # partial table below contains new columns and existing columns with property changes
        filters: List[Tuple[TSchemaContractEntities, str, TSchemaEvolutionMode]] = []
        # properties from TColumnType that are checked under data_type contract
        data_type_props = get_type_hints(TColumnType)
        for column_name, column in list(partial_table["columns"].items()):
            # dlt cols may always be added
            if is_dlt_table_or_column(column_name, self._dlt_tables_prefix):
                continue
            is_variant = column.get("variant", False)
            # check if column already exists to distinguish between new column vs property change
            existing_col = existing_table["columns"].get(column_name) if existing_table else None
            is_new_column = not existing_col or not is_complete_column(existing_col)

            # column contract applies only to genuinely new columns
            if column_mode != "evolve" and not is_variant and is_new_column:
                error_msg = (
                    f"Can't add table column `{column_name}` to table `{table_name}` because"
                    " `columns` are frozen."
                )

                if raise_on_freeze and column_mode == "freeze":
                    raise DataValidationError(
                        self.name,
                        table_name,
                        column_name,
                        "columns",
                        "freeze",
                        existing_table,
                        schema_contract,
                        data_item,
                        error_msg,
                    )
                # filter column with name below
                filters.append(("columns", column_name, column_mode))
                # pop the column
                partial_table["columns"].pop(column_name)

            # data_type contract applies to: variant columns and type property changes on
            # existing complete columns. only TColumnType properties trigger the check.
            has_type_change = False
            if not is_new_column and not is_variant and existing_col:
                for prop in data_type_props:
                    if prop in column and column[prop] != existing_col.get(prop):  # type: ignore[literal-required]
                        has_type_change = True
                        break

            if data_mode != "evolve" and (is_variant or has_type_change):
                if is_variant:
                    error_msg = (
                        f"Can't add variant column `{column_name}` for table `{table_name}`"
                        " because `data_types` are frozen."
                    )
                else:
                    error_msg = (
                        f"Can't evolve column type of `{column_name}` in table"
                        f" `{table_name}` because `data_types` are frozen."
                        f" Existing column: {existing_col}. Incoming column: {column}."
                    )
                if raise_on_freeze and data_mode == "freeze":
                    raise DataValidationError(
                        self.name,
                        table_name,
                        column_name,
                        "data_type",
                        "freeze",
                        existing_table,
                        schema_contract,
                        data_item,
                        error_msg,
                    )
                # filter column with name below
                filters.append(("columns", column_name, data_mode))
                # pop the column
                partial_table["columns"].pop(column_name)

        return partial_table, filters

    @staticmethod
    def expand_schema_contract_settings(
        settings: TSchemaContract, default: TSchemaContractDict = None
    ) -> TSchemaContractDict:
        """Expand partial or shorthand settings into full settings dictionary using `default` for unset entities"""
        if isinstance(settings, str):
            settings = TSchemaContractDict(tables=settings, columns=settings, data_type=settings)
        return cast(
            TSchemaContractDict, {**(default or DEFAULT_SCHEMA_CONTRACT_MODE), **(settings or {})}
        )

    def resolve_contract_settings_for_table(
        self, table_name: str, new_table_schema: TTableSchema = None
    ) -> TSchemaContractDict:
        """Resolve the exact applicable schema contract settings for the table `table_name`. `new_table_schema` is added to the tree during the resolution."""

        settings: TSchemaContract = {}
        if not is_dlt_table_or_column(table_name, self._dlt_tables_prefix):
            if new_table_schema:
                tables = copy(self._schema_tables)
                tables[table_name] = new_table_schema
            else:
                tables = self._schema_tables
            try:
                settings = utils.get_inherited_table_hint(tables, table_name, "schema_contract")
            except ValueError:
                settings = self._settings.get("schema_contract", {})

        # expand settings, empty settings will expand into default settings
        return Schema.expand_schema_contract_settings(settings)

    def update_table(
        self,
        partial_table: TPartialTableSchema,
        normalize_identifiers: bool = True,
        merge_compound_props: bool = True,
    ) -> TPartialTableSchema:
        """Adds or merges `partial_table` into the schema.

        Args:
            partial_table: Table schema to add or merge
            normalize_identifiers: If True, normalizes identifiers using schema naming convention
            merge_compound_props: If False, compound properties (see `is_compound_prop()` in schema utils)
                in partial_table replace rather than merge with those in existing table.

        Returns:
            The partial table that was applied (either the input or the generated diff)

        Raises:
            ParentTableNotFoundException: If parent table specified but not present in schema
        """
        parent_table_name = partial_table.get("parent")
        if normalize_identifiers:
            partial_table = utils.normalize_table_identifiers(partial_table, self.naming)

        table_name = partial_table["name"]
        # check if parent table present
        if parent_table_name is not None:
            if self._schema_tables.get(parent_table_name) is None:
                raise ParentTableNotFoundException(
                    self.name,
                    table_name,
                    parent_table_name,
                    "If you declared nested hints, make sure you added all intermediate tables,"
                    " including those for which you do not declare any hints.",
                )
        table = self._schema_tables.get(table_name)
        if table is None:
            self._schema_tables[table_name] = partial_table
        else:
            partial_table = utils.merge_table(self.name, table, partial_table, merge_compound_props)

        self.data_item_normalizer.extend_table(table_name)
        return partial_table

    def update_schema(self, schema: "Schema") -> None:
        """Updates this schema from an incoming schema. Normalizes identifiers after updating normalizers."""
        # pass normalizer config
        self._settings = deepcopy(schema.settings)
        # make shallow copy of normalizer settings
        self._configure_normalizers(copy(schema._normalizers_config))
        self.data_item_normalizer.extend_schema(extend_tables=False)
        self._compile_settings()
        # update all tables starting for parents and then nested tables in order
        tables = list(schema.tables.values())
        for table in tables:
            if not utils.is_nested_table(table):
                for chain_table in utils.get_nested_tables(schema._schema_tables, table["name"]):
                    self.update_table(chain_table)

    def unify_schemas(self, schemas: Sequence["Schema"]) -> "Schema":
        """Create a new schema containing tables from self and all given schemas.

        All schemas must use the same naming convention (type, case sensitivity,
        and max_length). Tables are merged without normalizing identifiers
        (assumed already normalized). Parent tables are added before children.

        Raises:
            IncompatibleSchemaException: when naming conventions differ
        """
        all_names = sorted([self.name] + [s.name for s in schemas])
        unified_name = utils.normalize_schema_name("u_" + "_".join(all_names))
        unified = self.clone(with_name=unified_name)
        for schema in schemas:
            if str(unified.naming) != str(schema.naming):
                raise IncompatibleSchemaException(
                    unified.name,
                    schema.name,
                    f"naming convention mismatch: `{unified.naming}` vs `{schema.naming}`",
                )
            for table in schema.tables.values():
                if not utils.is_nested_table(table):
                    for chain_table in utils.get_nested_tables(
                        schema._schema_tables, table["name"]
                    ):
                        unified.update_table(chain_table, normalize_identifiers=False)
        return unified

    def drop_tables(self, table_names: Sequence[str]) -> List[TTableSchema]:
        """Drops tables from the schema and returns the dropped tables. List of table names
        must contain all nested tables to tables being dropped.
        """
        result = []
        candidates = set()

        for table_name in table_names:
            if self.get_table(table_name) and table_name not in candidates:
                candidates.add(table_name)
                # also add table chain
                candidates.update(
                    t["name"] for t in utils.get_nested_tables(self._schema_tables, table_name)
                )
        # compare extension with original list
        if orphaned := candidates.difference(table_names):
            raise SchemaCorruptedException(
                self._schema_name,
                "A set tables to drop would leave orphaned tables. Please use consistent list of "
                f"table names in `drop_table`. Orphaned tabled: {orphaned}",
            )
        # final drop
        for table_name in table_names:
            if table_name in candidates:
                result.append(self._schema_tables.pop(table_name))
                self.data_item_normalizer.remove_table(table_name)

        return result

    def filter_row_with_hint(
        self, table_name: str, hint_type: TColumnDefaultHint, row: StrAny
    ) -> StrAny:
        rv_row: DictStrAny = {}
        column_prop: TColumnProp = utils.hint_to_column_prop(hint_type)
        try:
            table = self.get_table_columns(table_name, include_incomplete=True)
            for column_name in table:
                if column_name in row:
                    hint_value = table[column_name][column_prop]
                    if not utils.has_default_column_prop_value(column_prop, hint_value):
                        rv_row[column_name] = row[column_name]
        except KeyError:
            for k, v in row.items():
                if self._infer_hint(hint_type, k):
                    rv_row[k] = v

        # dicts are ordered and we will return the rows with hints in the same order as they appear in the columns
        return rv_row

    def merge_hints(
        self,
        new_hints: Mapping[TColumnDefaultHint, Sequence[TSimpleRegex]],
        replace: bool = False,
        normalize_identifiers: bool = True,
    ) -> None:
        """Merges or replace existing default hints with `new_hints`. Normalizes names in column regexes if possible. Compiles setting at the end

        NOTE: you can manipulate default hints collection directly via `Schema.settings` as long as you call Schema._compile_settings() at the end.
        """
        self._merge_hints(new_hints, replace=replace, normalize_identifiers=normalize_identifiers)
        self._compile_settings()

    def update_preferred_types(
        self,
        new_preferred_types: Mapping[TSimpleRegex, TDataType],
        normalize_identifiers: bool = True,
    ) -> None:
        """Updates preferred types dictionary with `new_preferred_types`. Normalizes names in column regexes if possible. Compiles setting at the end

        NOTE: you can manipulate preferred hints collection directly via `Schema.settings` as long as you call Schema._compile_settings() at the end.
        """
        self._update_preferred_types(new_preferred_types, normalize_identifiers)
        self._compile_settings()

    def add_type_detection(self, detection: TTypeDetections) -> None:
        """Add type auto detection to the schema."""
        if detection not in self.settings["detections"]:
            self.settings["detections"].append(detection)
            self._compile_settings()

    def remove_type_detection(self, detection: TTypeDetections) -> None:
        """Adds type auto detection to the schema."""
        if detection in self.settings["detections"]:
            self.settings["detections"].remove(detection)
            self._compile_settings()

    def get_new_table_columns(
        self,
        table_name: str,
        existing_columns: TTableSchemaColumns,
        case_sensitive: bool,
        include_incomplete: bool = False,
    ) -> List[TColumnSchema]:
        """Gets new columns to be added to `existing_columns` to bring them up to date with `table_name` schema.
        Columns names are compared case sensitive by default. `existing_column` names are expected to be normalized.
        Typically they come from the destination schema. Columns that are in `existing_columns` and not in `table_name` columns are ignored.

        Optionally includes incomplete columns (without data type)"""
        casefold_f: Callable[[str], str] = str.casefold if not case_sensitive else str
        casefold_existing = {
            casefold_f(col_name): col for col_name, col in existing_columns.items()
        }
        if len(existing_columns) != len(casefold_existing):
            raise SchemaCorruptedException(
                self.name,
                "A set of existing columns passed to `get_new_table_columns` table"
                f" `{table_name}` has colliding names when case insensitive comparison is used."
                f" Original names: {list(existing_columns.keys())}. Case-folded names:"
                f" {list(casefold_existing.keys())}",
            )
        diff_c: List[TColumnSchema] = []
        updated_columns = self.get_table_columns(table_name, include_incomplete=include_incomplete)
        for c in updated_columns.values():
            if casefold_f(c["name"]) not in casefold_existing:
                diff_c.append(c)
        return diff_c

    def get_table(self, table_name: str) -> TTableSchema:
        try:
            return self._schema_tables[table_name]
        except KeyError as k_exc:
            raise TableNotFound(self._schema_name, table_name) from k_exc

    def get_table_columns(
        self, table_name: str, include_incomplete: bool = False
    ) -> TTableSchemaColumns:
        """Gets columns of `table_name`. Optionally includes incomplete columns"""
        table = self.get_table(table_name)
        if include_incomplete:
            return table["columns"]
        else:
            return {k: v for k, v in table["columns"].items() if utils.is_complete_column(v)}

    def data_tables(
        self, seen_data_only: bool = False, include_incomplete: bool = False
    ) -> List[TTableSchema]:
        """Gets list of all tables, that hold the loaded data. Excludes dlt tables. Excludes incomplete tables (ie. without columns)"""
        return [
            t
            for t in self._schema_tables.values()
            if not is_dlt_table_or_column(t["name"], self._dlt_tables_prefix)
            and (
                (
                    include_incomplete
                    or len(self.get_table_columns(t["name"], include_incomplete)) > 0
                )
                and (not seen_data_only or utils.has_table_seen_data(t))
            )
        ]

    def data_table_names(
        self, seen_data_only: bool = False, include_incomplete: bool = False
    ) -> List[str]:
        """Returns list of table names. Excludes dlt table names."""
        return [
            t["name"]
            for t in self.data_tables(
                seen_data_only=seen_data_only, include_incomplete=include_incomplete
            )
        ]

    def dlt_tables(self) -> List[TTableSchema]:
        """Gets dlt tables"""
        return [
            t
            for t in self._schema_tables.values()
            if is_dlt_table_or_column(t["name"], self._dlt_tables_prefix)
        ]

    def dlt_table_names(self) -> List[str]:
        """Returns list of dlt table names."""
        return [t["name"] for t in self.dlt_tables()]

    def get_preferred_type(self, col_name: str) -> Optional[TDataType]:
        return next((m[1] for m in self._compiled_preferred_types if m[0].search(col_name)), None)

    def is_new_table(self, table_name: str) -> bool:
        """Returns true if this table does not exist OR is incomplete (has only incomplete columns) and therefore new"""
        return (table_name not in self.tables) or (
            not [
                c
                for c in self._schema_tables[table_name]["columns"].values()
                if utils.is_complete_column(c)
            ]
        )

    @property
    def version(self) -> int:
        """Version of the schema content that takes into account changes from the time of schema loading/creation.
        The stored version is increased by one if content was modified

        Returns:
            int: Current schema version
        """
        return utils.bump_version_if_modified(self.to_dict())[0]

    @property
    def stored_version(self) -> int:
        """Version of the schema content form the time of schema loading/creation.

        Returns:
            int: Stored schema version
        """
        return self._stored_version

    @property
    def version_hash(self) -> str:
        """Current version hash of the schema, recomputed from the actual content"""
        return utils.bump_version_if_modified(self.to_dict())[1]

    @property
    def previous_hashes(self) -> Sequence[str]:
        """Current version hash of the schema, recomputed from the actual content"""
        return utils.bump_version_if_modified(self.to_dict())[3]

    @property
    def stored_version_hash(self) -> str:
        """Version hash of the schema content form the time of schema loading/creation."""
        return self._stored_version_hash

    @property
    def is_modified(self) -> bool:
        """Checks if schema was modified from the time it was saved or if this is a new schema

        A current version hash is computed and compared with stored version hash
        """
        return self.version_hash != self._stored_version_hash

    @property
    def is_new(self) -> bool:
        """Checks if schema was ever saved"""
        return self._stored_version_hash is None

    @property
    def name(self) -> str:
        return self._schema_name

    @property
    def tables(self) -> TSchemaTables:
        """Dictionary of schema tables"""
        return self._schema_tables

    @property
    def references(self) -> list[TTableReferenceStandalone]:
        """References between tables"""
        all_references: list[TTableReferenceStandalone] = []
        for table_name, table in self.tables.items():
            # TODO more specific error handling than ValueError
            try:
                parent_ref = utils.create_parent_child_reference(self.tables, table_name)
                all_references.append(cast(TTableReferenceStandalone, parent_ref))
            except ValueError:
                pass

            try:
                root_ref = utils.create_root_child_reference(self.tables, table_name)
                all_references.append(cast(TTableReferenceStandalone, root_ref))
            except ValueError:
                pass

            try:
                load_table_ref = utils.create_load_table_reference(
                    self.tables[table_name], naming=self.naming
                )
                all_references.append(cast(TTableReferenceStandalone, load_table_ref))
            except ValueError:
                pass

            refs = table.get("references")
            if not refs:
                continue

            for ref in refs:
                top_level_ref: TTableReference = ref.copy()
                if top_level_ref.get("table") is None:
                    top_level_ref["table"] = table_name

                all_references.append(cast(TTableReferenceStandalone, top_level_ref))

        # internal references with `_dlt_version` need to be extracted once
        try:
            version_table_hash_ref = utils.create_version_and_loads_hash_reference(
                self.tables, naming=self.naming
            )
            version_table_schema_name_ref = utils.create_version_and_loads_schema_name_reference(
                self.tables, naming=self.naming
            )
            all_references.append(cast(TTableReferenceStandalone, version_table_hash_ref))
            all_references.append(cast(TTableReferenceStandalone, version_table_schema_name_ref))
        except ValueError:
            pass

        return all_references

    @property
    def settings(self) -> TSchemaSettings:
        return self._settings

    def __repr__(self) -> str:
        kwargs = {
            "name": self.name,
            "version": self.version,
            "tables": list(self.tables),
            "version_hash": self.version_hash,
        }
        return simple_repr("dlt.Schema", **without_none(kwargs))

    def _repr_html_(self, **kwargs: Any) -> str:
        """Render the Schema has a graphviz graph and display it using HTML

        This method is automatically called by notebooks renderers (IPython, marimo, etc.)
        ref: https://ipython.readthedocs.io/en/stable/config/integrating.html

        `dlt.helpers.graphviz.render_with_html()` has not external Python or system dependencies.
        """
        from dlt.helpers.graphviz import _render_dot_with_html

        return _render_dot_with_html(self.to_dot(**kwargs))

    def to_dict(
        self,
        remove_defaults: bool = False,
        remove_processing_hints: bool = False,
        bump_version: bool = True,
    ) -> TStoredSchema:
        stored_schema: TStoredSchema = {
            "version": self._stored_version,
            "version_hash": self._stored_version_hash,
            "engine_version": Schema.ENGINE_VERSION,
            "name": self._schema_name,
            "tables": self._schema_tables,
            "settings": self._settings,
            "normalizers": self._normalizers_config,
            "previous_hashes": self._stored_previous_hashes,
        }
        if self._imported_version_hash and not remove_defaults:
            stored_schema["imported_version_hash"] = self._imported_version_hash
        if self._schema_description:
            stored_schema["description"] = self._schema_description

        # remove processing hints that could be created by normalize and load steps
        if remove_processing_hints:
            stored_schema["tables"] = utils.remove_processing_hints(
                deepcopy(stored_schema["tables"])
            )

        # bump version if modified
        if bump_version:
            utils.bump_version_if_modified(stored_schema)
        # remove defaults after bumping version
        if remove_defaults:
            utils.remove_defaults(stored_schema)
        return stored_schema

    def to_pretty_json(
        self, remove_defaults: bool = True, remove_processing_hints: bool = False
    ) -> str:
        d = self.to_dict(
            remove_defaults=remove_defaults, remove_processing_hints=remove_processing_hints
        )
        return utils.to_pretty_json(d)

    def to_pretty_yaml(
        self, remove_defaults: bool = True, remove_processing_hints: bool = False
    ) -> str:
        d = self.to_dict(
            remove_defaults=remove_defaults, remove_processing_hints=remove_processing_hints
        )
        return utils.to_pretty_yaml(d)

    def to_dbml(
        self,
        remove_processing_hints: bool = False,
        include_dlt_tables: bool = True,
        include_internal_dlt_ref: bool = True,
        include_parent_child_ref: bool = True,
        include_root_child_ref: bool = True,
        group_by_resource: bool = False,
    ) -> str:
        from dlt.helpers.dbml import schema_to_dbml

        stored_schema = self.to_dict(
            # setting this to `True` removes `name` fields that are used in `schema_to_dbml()`
            # if required, we can refactor `dlt.helpers.dbml` to support this
            remove_defaults=False,
            remove_processing_hints=remove_processing_hints,
        )

        # NOTE `allow_custom_dbml_properties` is not exposed because it produces invalid DBML
        dbml_schema = schema_to_dbml(
            stored_schema,
            include_dlt_tables=include_dlt_tables,
            include_internal_dlt_ref=include_internal_dlt_ref,
            include_parent_child_ref=include_parent_child_ref,
            include_root_child_ref=include_root_child_ref,
            group_by_resource=group_by_resource,
        )
        return str(dbml_schema.dbml)

    def to_dot(
        self,
        remove_processing_hints: bool = False,
        include_dlt_tables: bool = True,
        include_internal_dlt_ref: bool = True,
        include_parent_child_ref: bool = True,
        include_root_child_ref: bool = True,
        group_by_resource: bool = False,
    ) -> str:
        """Convert schema to a Graphviz DOT string.

        Args:
            remove_processing_hints: If True, remove hints used for data processing and redundant information.
                This reduces the size of the schema and improves readability.
            include_dlt_tables: If True, include data tables and internal dlt tables. This will influence table
                references and groups produced.
            include_internal_dlt_ref: If True, include references between tables `_dlt_version`, `_dlt_loads` and `_dlt_pipeline_state`
            include_parent_child_ref: If True, include references from `child._dlt_parent_id` to `parent._dlt_id`
            include_root_child_ref: If True, include references from `child._dlt_root_id` to `root._dlt_id`
            group_by_resource: If True, group tables by resource and create subclusters.

        Returns:
            A DOT string of the schema
        """
        from dlt.helpers.graphviz import schema_to_graphviz

        stored_schema = self.to_dict(
            # setting this to `True` removes `name` fields that are used in `schema_to_dbml()`
            # if required, we can refactor `dlt.helpers.dbml` to support this
            remove_defaults=False,
            remove_processing_hints=remove_processing_hints,
        )

        dot = schema_to_graphviz(
            stored_schema,
            include_dlt_tables=include_dlt_tables,
            include_internal_dlt_ref=include_internal_dlt_ref,
            include_parent_child_ref=include_parent_child_ref,
            include_root_child_ref=include_root_child_ref,
            group_by_resource=group_by_resource,
        )
        return dot

    def to_mermaid(
        self,
        remove_processing_hints: bool = False,
        hide_columns: bool = False,
        hide_descriptions: bool = False,
        include_dlt_tables: bool = True,
    ) -> str:
        """Convert schema to a Mermaid diagram string.
        Args:
            remove_processing_hints: If True, remove hints used for data processing and redundant information.
                This reduces the size of the schema and improves readability.
            hide_columns: If True, the diagram hides columns details. This helps readability of large diagrams.
            hide_descriptions: If True, hide the column descriptions
            include_dlt_tables: If `True` (the default), internal dlt tables (`_dlt_version`,
                `_dlt_loads`, `_dlt_pipeline_state`)

        Returns:
            A string containing a Mermaid ERdiagram of the schema.
        """
        from dlt.helpers.mermaid import schema_to_mermaid

        stored_schema = self.to_dict(
            # setting this to `True` removes `name` fields that are used in `schema_to_dbml()`
            # if required, we can refactor `dlt.helpers.dbml` to support this
            remove_defaults=False,
            remove_processing_hints=remove_processing_hints,
        )

        return schema_to_mermaid(
            stored_schema,
            references=self.references,
            hide_columns=hide_columns,
            hide_descriptions=hide_descriptions,
            include_dlt_tables=include_dlt_tables,
        )

    def clone(
        self,
        with_name: str = None,
        remove_processing_hints: bool = False,
        update_normalizers: bool = False,
    ) -> "Schema":
        """Make a deep copy of the schema, optionally changing the name, removing processing markers and updating normalizers and identifiers in the schema if `update_normalizers` is True
        Processing markers are `x-` hints created by normalizer (`x-normalizer`) and loader (`x-loader`) to ie. mark newly inferred tables and tables that seen data.
        Note that changing of name will break the previous version chain
        """
        d = deepcopy(
            self.to_dict(bump_version=False, remove_processing_hints=remove_processing_hints)
        )
        if with_name is not None:
            d["version"] = d["version_hash"] = None
            d.pop("imported_version_hash", None)
            d["name"] = with_name
            d["previous_hashes"] = []
        schema = Schema.from_stored_schema(d)
        # update normalizers and possibly all schema identifiers
        if update_normalizers:
            schema.update_normalizers()
        return schema

    def update_normalizers(self) -> None:
        """Looks for new normalizer configuration or for destination capabilities context and updates all identifiers in the schema

        Table and column names will be normalized with new naming convention, except tables that have seen data ('x-normalizer`) which will
        raise if any identifier is to be changed.
        Default hints, preferred data types and normalize configs (ie. column propagation) are normalized as well. Regexes are included as long
        as textual parts can be extracted from an expression.
        """
        self._configure_normalizers(configured_normalizers(schema_name=self._schema_name))
        self.data_item_normalizer.extend_schema()
        self._compile_settings()

    def will_update_normalizers(self) -> bool:
        """Checks if schema has any pending normalizer updates due to configuration or destination capabilities"""

        # import desired modules
        _, to_naming, _ = import_normalizers(
            configured_normalizers(schema_name=self._schema_name), self._normalizers_config
        )
        return type(to_naming) is not type(self.naming)  # noqa

    def set_schema_contract(self, settings: TSchemaContract) -> None:
        if not settings:
            self._settings.pop("schema_contract", None)
        else:
            self._settings["schema_contract"] = settings

    def _infer_hint(self, hint_type: TColumnDefaultHint, col_name: str) -> bool:
        if hint_type in self._compiled_hints:
            return any(h.search(col_name) for h in self._compiled_hints[hint_type])
        else:
            return False

    def _merge_hints(
        self,
        new_hints: Mapping[TColumnDefaultHint, Sequence[TSimpleRegex]],
        replace: bool = False,
        normalize_identifiers: bool = True,
    ) -> None:
        """Used by `merge_hints method, does not compile settings at the end"""
        # validate regexes
        validate_dict(
            TSchemaSettings,
            {"default_hints": new_hints},
            ".",
            validator_f=utils.simple_regex_validator,
        )
        if normalize_identifiers:
            new_hints = self._normalize_default_hints(new_hints)
        # prepare hints to be added
        default_hints = self._settings.setdefault("default_hints", {})
        # add `new_hints` to existing hints
        for h, l in new_hints.items():
            if h in default_hints and not replace:
                extend_list_deduplicated(default_hints[h], l, utils.canonical_simple_regex)
            else:
                # set new hint type
                default_hints[h] = l  # type: ignore

    def _update_preferred_types(
        self,
        new_preferred_types: Mapping[TSimpleRegex, TDataType],
        normalize_identifiers: bool = True,
    ) -> None:
        # validate regexes
        validate_dict(
            TSchemaSettings,
            {"preferred_types": new_preferred_types},
            ".",
            validator_f=utils.simple_regex_validator,
        )
        if normalize_identifiers:
            new_preferred_types = self._normalize_preferred_types(new_preferred_types)
        preferred_types = self._settings.setdefault("preferred_types", {})
        # we must update using canonical simple regex
        canonical_preferred = {
            utils.canonical_simple_regex(rx): rx for rx in preferred_types.keys()
        }
        for new_rx, new_dt in new_preferred_types.items():
            canonical_new_rx = utils.canonical_simple_regex(new_rx)
            if canonical_new_rx not in canonical_preferred:
                preferred_types[new_rx] = new_dt
            else:
                preferred_types[canonical_preferred[canonical_new_rx]] = new_dt

    def _bump_version(self) -> Tuple[int, str]:
        """Computes schema hash in order to check if schema content was modified. In such case the schema ``stored_version`` and ``stored_version_hash`` are updated.

        Should not be used directly. The method ``to_dict`` will generate TStoredSchema with correct value, only once before persisting schema to storage.

        Returns:
            Tuple[int, str]: Current (``stored_version``, ``stored_version_hash``) tuple
        """
        self._stored_version, self._stored_version_hash, _, self._stored_previous_hashes = (
            utils.bump_version_if_modified(self.to_dict(bump_version=False))
        )
        return self._stored_version, self._stored_version_hash

    def _drop_version(self) -> None:
        """Stores first prev hash as stored hash and decreases numeric version"""
        if len(self.previous_hashes) == 0 or self._stored_version is None:
            self._stored_version = None
            self._stored_version_hash = None
        else:
            self._stored_version -= 1
            self._stored_version_hash = self._stored_previous_hashes.pop(0)

    def _add_standard_tables(self) -> None:
        self._schema_tables[self.version_table_name] = utils.normalize_table_identifiers(
            utils.version_table(), self.naming
        )
        self._schema_tables[self.loads_table_name] = utils.normalize_table_identifiers(
            utils.loads_table(), self.naming
        )

    def _add_standard_hints(self) -> None:
        default_hints = utils.default_hints()
        if default_hints:
            self._merge_hints(default_hints, normalize_identifiers=False)
        type_detections = utils.standard_type_detections()
        if type_detections:
            self._settings["detections"] = type_detections

    def _normalize_default_hints(
        self, default_hints: Mapping[TColumnDefaultHint, Sequence[TSimpleRegex]]
    ) -> Dict[TColumnDefaultHint, List[TSimpleRegex]]:
        """Normalizes the column names in default hints. In case of column names that are regexes, normalization is skipped"""
        return {
            hint: [utils.normalize_simple_regex_column(self.naming, regex) for regex in regexes]
            for hint, regexes in default_hints.items()
        }

    def _normalize_preferred_types(
        self, preferred_types: Mapping[TSimpleRegex, TDataType]
    ) -> Dict[TSimpleRegex, TDataType]:
        """Normalizes the column names in preferred types mapping. In case of column names that are regexes, normalization is skipped"""
        return {
            utils.normalize_simple_regex_column(self.naming, regex): data_type
            for regex, data_type in preferred_types.items()
        }

    def _verify_update_normalizers(
        self,
        normalizers_config: TNormalizersConfig,
        to_naming: NamingConvention,
        from_naming: NamingConvention,
    ) -> TSchemaTables:
        """Verifies if normalizers can be updated before schema is changed"""
        allow_ident_change = normalizers_config.get(
            "allow_identifier_change_on_table_with_data", False
        )

        def _verify_identifiers(table: TTableSchema, norm_table: TTableSchema) -> None:
            if not allow_ident_change:
                # make sure no identifier got changed in table
                if norm_table["name"] != table["name"]:
                    raise TableIdentifiersFrozen(
                        self.name,
                        table["name"],
                        to_naming,
                        from_naming,
                        f"Attempt to rename table to `{norm_table['name']}`.",
                    )
                # if len(norm_table["columns"]) != len(table["columns"]):
                #     print(norm_table["columns"])
                #     raise TableIdentifiersFrozen(
                #         self.name,
                #         table["name"],
                #         to_naming,
                #         from_naming,
                #         "Number of columns changed after normalization. Some columns must have"
                #         " merged.",
                #     )
                col_diff = set(norm_table["columns"].keys()).symmetric_difference(
                    table["columns"].keys()
                )
                if len(col_diff) > 0:
                    raise TableIdentifiersFrozen(
                        self.name,
                        table["name"],
                        to_naming,
                        from_naming,
                        f"Some columns got renamed to `{col_diff}`.",
                    )

        naming_changed = from_naming and type(from_naming) is not type(to_naming)
        if naming_changed:
            schema_tables = {}
            # check dlt tables
            schema_seen_data = any(
                utils.has_table_seen_data(t) for t in self._schema_tables.values()
            )
            # modify dlt tables using original naming
            orig_dlt_tables = [
                (self.version_table_name, utils.version_table()),
                (self.loads_table_name, utils.loads_table()),
                (self.state_table_name, utils.pipeline_state_table(add_dlt_id=True)),
            ]
            for existing_table_name, original_table in orig_dlt_tables:
                table = self._schema_tables.get(existing_table_name)
                # state table is optional
                if table:
                    table = copy(table)
                    # keep all attributes of the schema table, copy only what we need to normalize
                    table["columns"] = original_table["columns"]
                    norm_table = utils.normalize_table_identifiers(table, to_naming)
                    table_seen_data = utils.has_table_seen_data(norm_table)
                    if schema_seen_data:
                        _verify_identifiers(table, norm_table)
                    schema_tables[norm_table["name"]] = norm_table

            schema_seen_data = False
            for table in self.data_tables(include_incomplete=True):
                # TODO: when lineage is fully implemented we should use source identifiers
                # not `table` which was already normalized
                norm_table = utils.normalize_table_identifiers(table, to_naming)
                table_seen_data = utils.has_table_seen_data(norm_table)
                if table_seen_data:
                    _verify_identifiers(table, norm_table)
                schema_tables[norm_table["name"]] = norm_table
                schema_seen_data |= table_seen_data
            if schema_seen_data and not allow_ident_change:
                # if any of the tables has seen data, fail naming convention change
                # NOTE: this will be dropped with full identifier lineage. currently we cannot detect
                # strict schemas being changed to lax
                raise TableIdentifiersFrozen(
                    self.name,
                    "-",
                    to_naming,
                    from_naming,
                    "Schema contains tables that received data. As a precaution, changing naming"
                    " conventions is disallowed until full identifier lineage is implemented.",
                )
            # re-index the table names
            return schema_tables
        else:
            return self._schema_tables

    def _replace_and_apply_naming(
        self,
        normalizers_config: TNormalizersConfig,
        to_naming: NamingConvention,
        from_naming: NamingConvention,
    ) -> None:
        """Normalizes all identifiers in the schema in place according to `to_naming`"""
        self._schema_tables = self._verify_update_normalizers(
            normalizers_config, to_naming, from_naming
        )
        self._normalizers_config = normalizers_config
        self.naming = to_naming
        # name normalization functions
        self._dlt_tables_prefix = to_naming.normalize_table_identifier(DLT_NAME_PREFIX)
        self.version_table_name = to_naming.normalize_table_identifier(VERSION_TABLE_NAME)
        self.loads_table_name = to_naming.normalize_table_identifier(LOADS_TABLE_NAME)
        self.state_table_name = to_naming.normalize_table_identifier(PIPELINE_STATE_TABLE_NAME)
        # do a sanity check - dlt tables must start with dlt prefix
        for table_name in [self.version_table_name, self.loads_table_name, self.state_table_name]:
            if not is_dlt_table_or_column(table_name, self._dlt_tables_prefix):
                raise SchemaCorruptedException(
                    self.name,
                    f"A naming convention `{self.naming.name()}` mangles `_dlt` table prefix to"
                    f" '{self._dlt_tables_prefix}'. A table '{table_name}' does not start with it.",
                )
        # normalize default hints
        if default_hints := self._settings.get("default_hints"):
            self._settings["default_hints"] = self._normalize_default_hints(default_hints)
        # normalized preferred types
        if preferred_types := self.settings.get("preferred_types"):
            self._settings["preferred_types"] = self._normalize_preferred_types(preferred_types)

    def _configure_normalizers(self, explicit_normalizers: TNormalizersConfig) -> None:
        """Gets naming and item normalizer from schema yaml, config providers and destination capabilities and applies them to schema."""
        # preserve current schema settings if not explicitly set in `explicit_normalizers`
        if explicit_normalizers and self._normalizers_config:
            for prop_ in [
                "use_break_path_on_normalize",
                "allow_identifier_change_on_table_with_data",
            ]:
                if prop_ in self._normalizers_config and prop_ not in explicit_normalizers:
                    explicit_normalizers[prop_] = self._normalizers_config[prop_]  # type: ignore[literal-required]

        normalizers_config, to_naming, item_normalizer_class = import_normalizers(
            explicit_normalizers, self._normalizers_config
        )
        self._replace_and_apply_naming(normalizers_config, to_naming, self.naming)
        # data item normalization function
        self.data_item_normalizer = item_normalizer_class(self)

    def _reset_schema(self, name: str, normalizers: TNormalizersConfig = None) -> None:
        self._schema_tables: TSchemaTables = {}
        self._schema_name: str = None
        self._stored_version = None
        self._stored_version_hash: str = None
        self._imported_version_hash: str = None
        self._schema_description: str = None
        self._stored_previous_hashes: List[str] = []

        self._settings: TSchemaSettings = {}
        self._compiled_preferred_types: List[Tuple[REPattern, TDataType]] = []
        self._compiled_hints: Dict[TColumnDefaultHint, Sequence[REPattern]] = {}
        self._compiled_excludes: Dict[str, Sequence[REPattern]] = {}
        self._compiled_includes: Dict[str, Sequence[REPattern]] = {}
        self._type_detections: Sequence[TTypeDetections] = None

        self._normalizers_config = None
        self.naming = None
        self.data_item_normalizer = None

        # verify schema name - it does not depend on normalizers
        self._set_schema_name(name)
        # add standard hints
        self._add_standard_hints()
        # configure normalizers, including custom config if present
        if not normalizers:
            normalizers = configured_normalizers(schema_name=self._schema_name)
        self._configure_normalizers(normalizers)
        self.data_item_normalizer.extend_schema()  # type: ignore[attr-defined]
        # add version tables
        self._add_standard_tables()
        # compile all known regexes
        self._compile_settings()

    def _from_stored_schema(self, stored_schema: TStoredSchema) -> None:
        self._schema_tables = stored_schema.get("tables") or {}
        if self.version_table_name not in self._schema_tables:
            raise SchemaCorruptedException(
                stored_schema["name"], f"Schema must contain table `{self.version_table_name}`"
            )
        if self.loads_table_name not in self._schema_tables:
            raise SchemaCorruptedException(
                stored_schema["name"], f"Schema must contain table `{self.loads_table_name}`"
            )
        self._stored_version = stored_schema["version"]
        self._stored_version_hash = stored_schema["version_hash"]
        self._imported_version_hash = stored_schema.get("imported_version_hash")
        self._schema_description = stored_schema.get("description")
        self._settings = stored_schema.get("settings") or {}
        self._stored_previous_hashes = stored_schema.get("previous_hashes")
        self._compile_settings()

    def _set_schema_name(self, name: str) -> None:
        """Validates and sets the schema name. Validation does not need naming convention and checks against Python identifier"""
        if not utils.is_valid_schema_name(name):
            raise InvalidSchemaName(name)
        self._schema_name = name

    def _compile_settings(self) -> None:
        # if self._settings:
        for pattern, dt in self._settings.get("preferred_types", {}).items():
            # add tuples to be searched in coercions
            self._compiled_preferred_types.append((utils.compile_simple_regex(pattern), dt))
        for hint_name, hint_list in self._settings.get("default_hints", {}).items():
            # compile hints which are column matching regexes
            self._compiled_hints[hint_name] = list(map(utils.compile_simple_regex, hint_list))
        if self._schema_tables:
            for table in self._schema_tables.values():
                if "filters" in table:
                    if "excludes" in table["filters"]:
                        self._compiled_excludes[table["name"]] = list(
                            map(utils.compile_simple_regex, table["filters"]["excludes"])
                        )
                    if "includes" in table["filters"]:
                        self._compiled_includes[table["name"]] = list(
                            map(utils.compile_simple_regex, table["filters"]["includes"])
                        )
        # look for auto-detections in settings and then normalizer
        self._type_detections = self._settings.get("detections") or self._normalizers_config.get("detections") or []  # type: ignore

    def __getstate__(self) -> Any:
        state = self.__dict__.copy()
        del state["naming"]
        del state["data_item_normalizer"]
        return state



from typing import IO, Any, Union
import orjson

from dlt.common.json import (
    custom_pua_encode,
    custom_pua_decode_nested,
    custom_encode,
    set_custom_encoder_impl,
    TPuaDecoders,
    DECODERS,
)
from dlt.common.typing import AnyFun

_impl_name = "orjson"

set_custom_encoder = set_custom_encoder_impl


def _dumps(
    obj: Any, sort_keys: bool, pretty: bool, default: AnyFun = custom_encode, options: int = 0
) -> bytes:
    options = options | orjson.OPT_UTC_Z | orjson.OPT_NON_STR_KEYS
    if pretty:
        options |= orjson.OPT_INDENT_2
    if sort_keys:
        options |= orjson.OPT_SORT_KEYS
    return orjson.dumps(obj, default=default, option=options)


def dump(obj: Any, fp: IO[bytes], sort_keys: bool = False, pretty: bool = False) -> None:
    fp.write(_dumps(obj, sort_keys, pretty))


def typed_dump(obj: Any, fp: IO[bytes], pretty: bool = False) -> None:
    fp.write(typed_dumpb(obj, pretty=pretty))


def typed_dumpb(obj: Any, sort_keys: bool = False, pretty: bool = False) -> bytes:
    return _dumps(obj, sort_keys, pretty, custom_pua_encode, orjson.OPT_PASSTHROUGH_DATETIME)


def typed_dumps(obj: Any, sort_keys: bool = False, pretty: bool = False) -> str:
    return typed_dumpb(obj, sort_keys, pretty).decode("utf-8")


def typed_loads(s: str) -> Any:
    return custom_pua_decode_nested(loads(s))


def typed_loadb(s: Union[bytes, bytearray, memoryview], decoders: TPuaDecoders = DECODERS) -> Any:
    return custom_pua_decode_nested(loadb(s), decoders)


def dumps(obj: Any, sort_keys: bool = False, pretty: bool = False) -> str:
    return _dumps(obj, sort_keys, pretty).decode("utf-8")


def dumpb(obj: Any, sort_keys: bool = False, pretty: bool = False) -> bytes:
    return _dumps(obj, sort_keys, pretty)


def load(fp: IO[bytes]) -> Any:
    return orjson.loads(fp.read())


def loads(s: str) -> Any:
    return orjson.loads(s.encode("utf-8"))


def loadb(s: Union[bytes, bytearray, memoryview]) -> Any:
    return orjson.loads(s)



import sys
import queue
from contextlib import contextmanager
from subprocess import PIPE, CalledProcessError
from threading import Thread
from typing import Any, Generator, Iterator, List, Tuple, Literal

from dlt.common.runners.venv import Venv
from dlt.common.runners.synth_pickle import decode_obj, decode_last_obj, encode_obj
from dlt.common.typing import AnyFun

# file number of stdout (1) and stderr (2)
OutputStdStreamNo = Literal[1, 2]


@contextmanager
def exec_to_stdout(f: AnyFun) -> Iterator[Any]:
    """Executes parameter-less function f and encodes the pickled return value to stdout. In case of exceptions, encodes the pickled exceptions to stderr"""
    rv: Any = None
    try:
        rv = f()
        yield rv
    except Exception as ex:
        print(encode_obj(ex), file=sys.stderr, flush=True)  # noqa
        raise
    finally:
        if rv is not None:
            print(encode_obj(rv), flush=True)  # noqa


def iter_std(
    venv: Venv, command: str, *script_args: Any
) -> Iterator[Tuple[OutputStdStreamNo, str]]:
    """Starts a process `command` with `script_args` in environment `venv` and returns iterator
    of (filno, line) tuples where `fileno` is 1 for stdout and 2 for stderr. `line` is
    a content of a line with stripped new line character.

    Use -u in scripts_args for unbuffered python execution
    """
    with venv.start_command(
        command,
        *script_args,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=1,
        text=True,
        errors="backslashreplace",
    ) as process:
        exit_code: int = None
        q_: queue.Queue[Tuple[OutputStdStreamNo, str]] = queue.Queue()

        def _r_q(std_: OutputStdStreamNo) -> None:
            stream_ = process.stderr if std_ == 2 else process.stdout
            for line in iter(stream_.readline, ""):
                q_.put((std_, line.rstrip("\n")))
            # close queue
            q_.put(None)

        # read stderr with a thread, selectors do not work on windows
        t_out = Thread(target=_r_q, args=(1,), daemon=True)
        t_out.start()
        t_err = Thread(target=_r_q, args=(2,), daemon=True)
        t_err.start()
        while line := q_.get():
            yield line

        # get exit code
        exit_code = process.wait()
        # wait till stderr is received
        t_out.join()
        t_err.join()

        # we fail iterator if exit code is not 0
        if exit_code != 0:
            raise CalledProcessError(exit_code, command, output="", stderr="")


def iter_stdout(venv: Venv, command: str, *script_args: Any) -> Iterator[str]:
    # start a process in virtual environment, assume that text comes from stdout
    with venv.start_command(
        command,
        *script_args,
        stdout=PIPE,
        stderr=PIPE,
        bufsize=1,
        text=True,
        errors="backslashreplace",
    ) as process:
        exit_code: int = None
        line = ""
        stderr: List[str] = []

        def _r_stderr() -> None:
            nonlocal stderr
            for line in iter(process.stderr.readline, ""):
                stderr.append(line)

        # read stderr with a thread, selectors do not work on windows
        t = Thread(target=_r_stderr, daemon=True)
        t.start()

        # read stdout with
        for line in iter(process.stdout.readline, ""):
            yield line.rstrip("\n")

        # get exit code
        exit_code = process.wait()
        # wait till stderr is received
        t.join()

        # we fail iterator if exit code is not 0
        if exit_code != 0:
            raise CalledProcessError(exit_code, command, output=line, stderr="".join(stderr))


def iter_stdout_with_result(
    venv: Venv, command: str, *script_args: Any
) -> Generator[str, None, Any]:
    """Yields stdout lines coming from remote process and returns the last result decoded with decode_obj. In case of exit code != 0 if exception is decoded
    it will be raised, otherwise CalledProcessError is raised"""
    last_result: Any = None
    try:
        for line in iter_stdout(venv, command, *script_args):
            # attempt to decode line
            result = decode_obj(line, ignore_pickle_errors=True)
            # keep last decoded result
            if result is not None:
                last_result = result
            else:
                # yield other lines
                yield line
        return last_result
    except CalledProcessError as cpe:
        # try to find last object in stderr
        if cpe.stderr:
            # if exception was decoded from stderr
            exception = decode_last_obj(cpe.stderr.split("\n"), ignore_pickle_errors=False)
            if isinstance(exception, Exception):
                raise exception from cpe
            else:
                sys.stderr.write(cpe.stderr)
        # otherwise reraise cpe
        raise
